from .utils import (grouper, multiprocess, find_files, file_transaction,
                    safe_makedir, file_exists, run, cp, mv, logger, tmp_dir,
                    gunzip_file, pigz_file)
from .fastx import (readfx, complexity_filter, read_count, print_fasta_record,
                    munge_header, length_filter, fasta_stats, length_sort,
                    remove_seq_wraps, gc_skew_and_content, trim_fastx,
                    check_sync, tmp_split_reads)
from .apps import (index_bam, run_fastqc, trimmomatic,
                   trimmomatic_version, run_kmernorm, kmernorm_version, spades,
                   spades_version, parse_bcl2fastq_log, bwa_mem, blastn,
                   blast_formatter, tetramer_pca, reference_filter,
                   blast_results_filter, run_checkm, run_seqtk_sample)
from .classifier import classify_by_ssu
# from .plot import



from __future__ import print_function
import csv
import fileinput
import logging
import os
import os.path as op
import random
import re
import shutil
import six
import tempfile
from contextlib import contextmanager
from interlap import InterLap
from itertools import groupby
from toolshed import nopen, reader

# local imports
from .fastx import (
    length_sort,
    munge_header,
    tmp_split_reads,
    readfx,
    check_sync,
    print_fasta_record,
)
from .utils import (
    file_exists,
    file_transaction,
    run,
    safe_makedir,
    grouper,
    tmp_dir,
    gunzip_file,
    pigz_file,
    touch,
)


logger = logging.getLogger(__name__)


def index_bam(bam_file):
    """
    Build an index for a bam file.

    parameters
        bam_file : alignment file path

    returns
        index file name : string
    """
    bam_index = bam_file + ".bai"
    if not file_exists(bam_index):
        with file_transaction(bam_index) as tx_out_file:
            run("samtools index %s %s" % (bam, tx_out_file))
    return bam_index


def _remove_option(options, item, flag=False):
    """
    remove item from options. item is assumed to not be a flag, therefore
    two sequential args are removed from options.

    parameters
        options : list of args
        item : string to search for among options
        flag : whether for not item is boolean

    returns
        options : list
    """
    if item in options:
        x = options.index(item)
        options.pop(x)
        if not flag:
            options.pop(x)
    return options


def filter_options(options, predefined_options):
    """
    filter predefined options from user-specified options.

    parameters
        options : string or list of options for application
        predefined_options : list of tuples specifying option and whether or
            not the option is a boolean flag

    returns
        options : list
    """
    options = options.split() if isinstance(options, six.string_types) else options
    for option, flag in predefined_options:
        options = _remove_option(options, option, flag)
    return options


def run_fastqc(fastq, out_dir, cores=1):
    """
    runs fastqc on fastq or fastqs as a space-delimited single string.

    parameters
        fastq : single file path or list of file paths
        out_dir : directory to write output files
        cores : cpu cores to utilize

    returns
        output directory : string
    """
    if not op.exists(out_dir):
        safe_makedir(out_dir)
    if isinstance(fastq, six.string_types):
        fastq = [fastq]
    for fq in fastq:
        fname, ext = op.splitext(op.basename(fq))
        if ext == ".gz":
            # fastqc only strips ".fastq"; variants like .fq will remain
            tfname, ext = op.splitext(fname)
            if ext == ".fastq":
                fname = tfname
        fname += "_fastqc.zip"
        out_file = op.join(out_dir, fname)
        if file_exists(out_file):
            continue
        with file_transaction(out_file) as tx:
            run(
                "fastqc --outdir %s --noextract -f fastq -t %d -q %s"
                % (op.dirname(tx), cores, fq),
                description="Running FastQC",
            )
    return out_dir


def trimmomatic(fastq, out_file, options, cores=1):
    """
    runs trimmomatic and returns path of out_file. if your reads are paired-end,
    fastq should be a list of [R1 path, R2 path].

    parameters
        fastq : file path or list of R1, R2 file paths
        out_file : file path if single-end, list otherwise (trimmed_R1,
            trimmed_orphan_R1, trimmed_R2, trimmed_orphan_R2)
        options : trimmomatic options

    returns
        output files : list
    """
    predefined_options = [
        ("-threads", False),
        ("-trimlog", False),
        ("PE", True),
        ("SE", True),
    ]
    if isinstance(fastq, six.string_types):
        fastq = [fastq]
    if len(fastq) > 1:
        assert len(out_file) == 4
    if file_exists(out_file):
        return out_file
    options = filter_options(options, predefined_options)
    flag_options = []
    trimmers = []
    for option in options:
        if option.startswith("-"):
            flag_options.append(option)
        else:
            trimmers.append(option)
    flag_options.extend(["-threads", str(cores)])

    with file_transaction(out_file) as tx_out_file:
        if len(fastq) == 1:
            cmd = "trimmomatic.sh SE {opts} {fastq} {out} {trimmers}".format(
                opts=" ".join(flag_options),
                fastq=fastq[0],
                out=tx_out_file,
                trimmers=" ".join(trimmers),
            )
        else:
            cmd = "trimmomatic.sh PE {opts} {fastq} {out} {trimmers}".format(
                opts=" ".join(flag_options),
                fastq=" ".join(fastq),
                out=" ".join(tx_out_file),
                trimmers=" ".join(trimmers),
            )
        run(cmd, description="Trimming")
    return out_file


def trimmomatic_version():
    """returns the version string for Trimmomatic."""
    return "Trimmomatic v0.32"


def run_kmernorm(fastq, out_file, options, cores=1):
    """
    perform digital normalization using kmernorm. unzip input file if gzipped
    and gzip input file after kmernorm.

    parameters
        fastq : file path; gzipped input will be decompressed and recompressed
        out_file : file path; file ending with '.gz' will be compressed
        options : kmernorm options
        cores : cpu cores to utilize

    returns
        normalized file path and surviving read pairs : tuple (string, int)
    """
    if file_exists(out_file):
        return out_file

    gzip_result = True if out_file.endswith(".gz") else False
    out_file = out_file.rsplit(".gz", 1)[0]
    if fastq.endswith(".gz"):
        fastq = gunzip_file(fastq)
    with file_transaction(out_file) as tx:
        run(
            "kmernorm %s %s > %s" % (options, fastq, tx),
            description="Normalizing with kmernorm",
        )
    if gzip_result:
        out_file = pigz_file(out_file, cores)
    # compress input file regardless
    fastq = pigz_file(fastq, cores)
    return out_file


def kmernorm_version():
    """
    Executes `kmernorm` and parses output for version info.

    returns
        version : string
    """
    version = ""
    for line in run("kmernorm 2>&1", iterable=True):
        line = line.strip()
        # 'Version' is misspelled in the software.
        if line.startswith("Ver"):
            line = line.split()
            version = line[1]
            break
    return "kmernorm " + version


def spades(sample, fastq, out_file, options, cores=1):
    """
    runs spades assembler on interleaved, PE fastq.

    parameters
        sample : name of sample being processed; will be prepended to contig name
        fastq : file path
        out_file : file path of assembled contigs
        options : spades options
        cores : cpu cores to utilize

    returns
        assembled contigs file path
    """
    if file_exists(out_file):
        return out_file
    predefined_options = [
        ("-o", False),
        ("--pe1-12", False),
        ("--12", False),
        ("-t", False),
        ("--threads", False),
    ]
    options = filter_options(options, predefined_options)
    with file_transaction(out_file) as tx:
        tx_dir = op.dirname(tx)
        contigs = tx_dir + "/contigs.fasta"
        cmd = "spades.py --12 {fq} --threads {cores} -o {out} {options}".format(
            fq=fastq, cores=cores, out=tx_dir, options=" ".join(options)
        )
        try:
            run(cmd)
        except:
            return out_file

        # sort the contigs
        sorted_contigs = os.path.join(tx_dir, sample + "_all_sorted_contigs.fasta")
        sorted_contigs = length_sort(contigs, sorted_contigs)
        # prepend sample name to contig header
        renamed_contigs = op.join(tx_dir, sample + "_all_contigs.fasta")
        renamed_contigs = munge_header(
            sorted_contigs, renamed_contigs, sample, intent="prepend-and-split", sep="_"
        )

        # need to rename spades output to match out_file
        shutil.move(renamed_contigs, tx)

    return out_file


def spades_version():
    """
    Executes `spades.py` and parses output for version info.

    returns
        version : string
    """
    version = ""
    for line in run("spades.py 2>&1 | head -1", iterable=True):
        version = line.rstrip("\r\n")
        break
    return version


def parse_bcl2fastq_log(log):
    """
    parses bcl2fastq log generated from bcl2fastq.py.

    returns
        version and run parameters : tuple (string, string)
    """
    version = ""
    parameters = ""
    if file_exists(log):
        for i, line in enumerate(nopen(log)):
            if i < 5 and line.startswith("bcl2fastq"):
                version = line.strip("\r\n")
                continue
            if "Command-line invocation" in line:
                parameters = (
                    '"%s"' % line.strip("\r\n").rsplit(": bcl2fastq", 1)[1].strip()
                )
            if version and parameters:
                break
    version = version if version else "Unable to detect"
    parameters = parameters if parameters else "Unable to detect"
    return version, parameters


def bwa_mem(fastq, out_file, index, options, cores=1):
    """
    align reads using bwa mem.

    parameters
        fastq : path to reads
        out_file : path to aligned reads bam
        index : path to bwa index
        options : bwa mem options
        cores : int

    returns
        output file path : string
    """
    if file_exists(out_file):
        return out_file
    predefined_options = [("-t", False)]
    options = filter_options(options, predefined_options)
    logger.info("Mapping %s using bwa mem" % bam)

    # TODO bwa index...
    with file_transaction(out_file) as tx_out_file:
        cmd = (
            "bwa mem -t {cores} {options} {index} {fastq} | samtools view "
            "-ShuF4q2 - | samtools sort -o -m 8G - tmp > {result}"
        ).format(
            cores=cores,
            options=" ".join(options),
            index=index,
            fastq=fastq,
            result=tx_out_file,
        )
        run(cmd)
        index_bam(tx_out_file)

    return out_file


def blastn(fasta, out_file, blast_db, options, cores=1):
    """
    align sequences using blastn, creating blast archive format (ASN1).

    fasta : file path as string
    out_file : result file path with extension ASN1
    options : blastn options except for db, num_threads, query, out, and outfmt

    returns ASN1 result path
    """
    if file_exists(out_file):
        return out_file
    predefined_options = [
        ("-num_threads", False),
        ("-out", False),
        ("-query", False),
        ("-outfmt", False),
        ("-db", False),
    ]
    options = filter_options(options, predefined_options)
    logger.info("Aligning %s using blastn" % fasta)

    with file_transaction(out_file) as tx_out_file:
        cmd = (
            "blastn -db {db} -outfmt 11 {options} -num_threads {cores} "
            "-out {result} -query {fasta}"
        ).format(
            db=blast_db,
            options=" ".join(options),
            cores=cores,
            result=tx_out_file,
            fasta=fasta,
        )
        run(cmd)

    return out_file


def blastn_fastq(fastq, out_file, blast_db, options, cores=1):
    """Align FASTQ sequences using blastn, creating a blast ASN1 archive.

    bioawk -c fastx '{print ">"$name"\n"$seq}' AD-782-J14_R1.fastq.gz | blastn -db nt -outfmt 11 -num_threads 80

    Args:
        fastq (str): fastq file path
        out_file (str): out file path to ASN1 archive
        blast_db (str): path or name of blast db
        options (str): blast options as you define them on the command line
        cores (Optional[int]): number of threads used by blast

    Returns:
        str

    Note:
        Adds `bioawk` dependency that should be removed in the future.

    """
    if file_exists(out_file):
        return out_file
    predefined_options = [
        ("-num_threads", False),
        ("-out", False),
        ("-query", False),
        ("-outfmt", False),
        ("-db", False),
    ]
    options = filter_options(options, predefined_options)
    logger.info("Aligning %s using blastn" % fastq)

    with file_transaction(out_file) as tx_out_file:
        cmd = r"""bioawk -c fastx '{print ">"$name"\n"$seq}' """
        cmd += (
            "{fastq} | blastn -db {db} -outfmt 11 {options} "
            "-num_threads {cores} -out {result} "
        ).format(
            fastq=fastq,
            db=blast_db,
            options=" ".join(options),
            cores=cores,
            result=tx_out_file,
        )
        run(cmd)

    return out_file


def blast_formatter(archive, out_file, options, out_fmt=None):
    """"""
    if file_exists(out_file):
        return out_file
    # gzipped somewhere outside, but after this method and the pipeline
    # is being re-run
    if file_exists(out_file + ".gz"):
        return out_file + ".gz"

    predefined_options = [("-archive", False), ("-outfmt", False)]
    options = filter_options(options, predefined_options)

    logger.info("Converting %s to %s" % (archive, out_file))

    blast_header = {
        "qseqid": "Query Seq-id",
        "qgi": "Query GI",
        "qacc": "Query accesion",
        "qaccver": "Query accesion.version",
        "qlen": "Query sequence length",
        "sseqid": "Subject Seq-id",
        "sallseqid": "All subject Seq-id(s)",
        "sgi": "Subject GI",
        "sallgi": "All subject GIs",
        "sacc": "Subject accession",
        "saccver": "Subject accession.version",
        "sallacc": "All subject accessions",
        "slen": "Subject sequence length",
        "qstart": "Start of alignment in query",
        "qend": "End of alignment in query",
        "sstart": "Start of alignment in subject",
        "send": "End of alignment in subject",
        "qseq": "Aligned part of query sequence",
        "sseq": "Aligned part of subject sequence",
        "evalue": "Expect value",
        "bitscore": "Bit score",
        "score": "Raw score",
        "length": "Alignment length",
        "pident": "Percentage of identical matches",
        "nident": "Number of identical matches",
        "mismatch": "Number of mismatches",
        "positive": "Number of positive-scoring matches",
        "gapopen": "Number of gap openings",
        "gaps": "Total number of gaps",
        "ppos": "Percentage of positive-scoring matches",
        "frames": "Query and subject frames",
        "qframe": "Query frame",
        "sframe": "Subject frame",
        "btop": "Blast traceback operations (BTOP)",
        "staxids": "unique Subject Taxonomy ID(s)",
        "sscinames": "unique Subject Scientific Name(s)",
        "scomnames": "unique Subject Common Name(s)",
        "sblastnames": "unique Subject Blast Name(s)",
        "sskingdoms": "unique Subject Super Kingdom(s)",
        "stitle": "Subject Title",
        "salltitles": "All Subject Title(s)",
        "sstrand": "Subject Strand",
        "qcovs": "Query Coverage Per Subject",
        "qcovhsp": "Query Coverage Per HSP",
    }

    with file_transaction(out_file) as tx_out_file:
        if out_fmt:
            n = out_fmt[0]
            if n == "6" or n == "10":
                hdr = [blast_header[k] for k in out_fmt[1:].split()]
                if n == "6":
                    sep = "\t"
                else:
                    sep = ","
                with open(tx_out_file, "w") as fh:
                    print(*hdr, sep=sep, file=fh)

            cmd = (
                "blast_formatter -archive {archive} {options} "
                "-outfmt '{fmt}' >> {result}"
            ).format(
                archive=archive,
                options=" ".join(options),
                fmt=out_fmt,
                result=tx_out_file,
            )
        else:
            cmd = (
                "blast_formatter -archive {archive} {options} " ">> {result}"
            ).format(archive=archive, options=" ".join(options), result=tx_out_file)
        run(cmd)
    return out_file


def tetramer_pca(fasta, out_dir, script_path, options, cores=1):
    sample_id = os.path.basename(fasta).rpartition(".")[0]
    # out_dir/sample_id + -<window size>-<step size>-PC.pdf
    out_files = [
        os.path.join(out_dir, sample_id + "-outliers.fasta"),
        os.path.join(out_dir, sample_id + "-outliers.xml"),
        os.path.join(out_dir, sample_id + "-tetramer-counts.csv"),
        os.path.join(out_dir, sample_id + "-tetramer-fail.csv"),
        os.path.join(out_dir, sample_id + "-tetramer-loading.csv"),
        os.path.join(out_dir, sample_id + "-tetramer-PC.csv"),
    ]

    if file_exists(out_files):
        return out_dir

    predefined_options = [
        ("--input", False),
        ("--output_dir", False),
        ("--num_threads", False),
    ]
    options = filter_options(options, predefined_options)

    logger.info("Running Tetramer PCA on %s" % fasta)

    # out_files = []
    # with file_transaction(out_files) as tx_out_files:
    cmd = (
        "Rscript --vanilla {script} --input {fasta} "
        "--output_dir {out_dir} --num_threads {cores} "
        "{options}"
    ).format(
        script=script_path,
        fasta=fasta,
        out_dir=out_dir,
        cores=cores,
        options=" ".join(options),
    )
    try:
        run(cmd, description="Running Tetramer PCA")
    except:
        return False
    return out_dir


@contextmanager
def bwa_index(reference, options=None):
    """Builds an index using `bwa index`.

    Args:
        reference (str): file path of reference fasta
        options (Optional[str]): options passed to bwa index, eg. "-a is"

    Yields:
        str: file path of reference as it's used as the prefix in `bwa index`
    """
    ref = op.abspath(reference)
    idx_files = [ref + x for x in [".amb", ".ann", ".bwt", ".pac", ".sa"]]
    if not file_exists(idx_files):
        logger.debug("Creating BWA index for %s" % ref)
        cmd = "bwa index {options} {ref}".format(
            options=options if options else "", ref=ref
        )
        run(cmd, "Building BWA index")
    yield reference


def _aln(ref, fastq, tmp="/tmp", threads=1, threshold=0.05):
    sai = os.path.join(tmp, "%09d.sai" % random.randrange(0, 1e10))
    with file_transaction(sai) as tx:
        cmd = ("bwa aln -n {threshold} -t {threads} " "{ref} {fastq} > {tx}").format(
            **locals()
        )
        run(cmd)
    return sai


def _sampe(ref, reads, tmp="/tmp", threads=1, threshold=0.05):
    r1, r2 = tmp_split_reads(reads, tmp)
    with bwa_index(ref) as bwaidx:
        r1_sai = _aln(bwaidx, r1, tmp, threads, threshold)
        r2_sai = _aln(bwaidx, r2, tmp, threads, threshold)
        sampe = (
            "bwa sampe {ref} {r1_sai} {r2_sai} {r1} {r2} " "| samtools view -SF0x0004 -"
        ).format(ref=bwaidx, r1_sai=r1_sai, r2_sai=r2_sai, r1=r1, r2=r2)
        for a in run(sampe, iterable=True):
            yield a


def reference_filter(fastq, fasta, out_file, hits_file=None, threshold=0.05, cores=1):
    """Align FASTQ file against a reference FASTA file generating a filtered
    FASTQ file. Optionally, if a file path is specified for alignment hits,
    reads are removed from FASTQ and written to hits.

    Args:
        fastq (str): path to interleaved fastq file
        fasta (str): path to reference fasta file
        out_file (str): path of output FASTQ with hits removed
        hits_file (Optional[str]): path of output FASTQ containing contamination hits only
        threshold (Optional[float]): allowable mismatch threshold of mapped reads
        cores (Optional[int]): bwa processing threads

    Returns:
        str or list: output file path or output file path and hits file path
    """
    if not hits_file and file_exists(out_file):
        return out_file
    # okay to have a 0 byte hits_file
    if hits_file and os.path.exists(hits_file) and file_exists(out_file):
        return out_file, hits_file

    logger.info("Filtering %s based on %s" % (fastq, fasta))
    if hits_file:
        out_files = [out_file, hits_file]
    else:
        out_files = [out_file]

    # file_transaction will yield a list when hits_file is used
    with file_transaction(out_files) as tx_files:
        hits = set()
        observed = 0
        survived = 0

        # this makes things easier if no hits_file is specified
        if isinstance(tx_files, six.string_types):
            tx_files = [tx_files]
        txdir = os.path.dirname(tx_files[0])

        for hit in _sampe(fasta, fastq, txdir, cores, threshold):
            hits.add(hit.partition("\t")[0])

        if len(hits) == 0:
            # still want to have an intermediate file here
            logger.info("No observed contamination.")
            # need to add .gz on output file if input file also has it
            shutil.copy(fastq, tx_files[0])
            if hits_file:
                touch(tx_files[1])
                return out_files
            else:
                return out_files[0]

        if hits_file:
            with open(tx_files[0], "w") as fq_fh, open(tx_files[1], "w") as ht_fh:
                for (n1, s1, q1), (n2, s2, q2) in grouper(2, readfx(fastq)):
                    check_sync(n1, n2)
                    observed += 1
                    # bwa strips read number from name so this works on R1 and R2
                    read_name = n1.partition("/")[0]
                    if read_name in hits:
                        print(
                            "@" + n1,
                            s1,
                            "+",
                            q1,
                            "@" + n2,
                            s2,
                            "+",
                            q2,
                            sep="\n",
                            file=ht_fh,
                        )
                    else:
                        survived += 1
                        print(
                            "@" + n1,
                            s1,
                            "+",
                            q1,
                            "@" + n2,
                            s2,
                            "+",
                            q2,
                            sep="\n",
                            file=fq_fh,
                        )
        else:
            with open(tx_files[0], "w") as fq_fh:
                for (n1, s1, q1), (n2, s2, q2) in grouper(2, readfx(fastq)):
                    check_sync(n1, n2)
                    observed += 1
                    # bwa strips read number from name so this works on R1 and R2
                    read_name = n1.partition("/")[0]
                    if read_name in hits:
                        continue
                    survived += 1
                    print(
                        "@" + n1,
                        s1,
                        "+",
                        q1,
                        "@" + n2,
                        s2,
                        "+",
                        q2,
                        sep="\n",
                        file=fq_fh,
                    )

        logger.info("Pairs observed: %d" % observed)
        logger.info("Pairs survived: %d" % survived)

        if tx_files[0].endswith(".gz"):
            pigz_file(tx_files[0], cores)
        if hits_file and tx_files[1].endswith(".gz"):
            pigz_file(tx_files[1], cores)

        count_file = tx_files[0].rsplit(".gz")[0] + ".count"
        with open(count_file, "w") as fh:
            print(survived * 2, file=fh)

    if hits_file:
        return out_files
    else:
        return out_file


def overlap_distance(coords):
    """
    >>> coords = [(1,3),(2,6),(8,9)]
    >>> overlap_distance(coords)
    8
    """
    overlap = InterLap()
    joined_intervals = InterLap()
    coords.sort(key=lambda x: x[0])
    overlap.add(coords)
    seen = set()
    for iset in coords:
        if iset in seen:
            continue
        seen.add(iset)
        start = iset[0]
        end = iset[1]
        for jset in overlap.find(iset):
            if jset in seen:
                continue
            seen.add(jset)
            start = min(start, jset[0])
            end = max([end, jset[1]])
        joined_intervals.add((start, end))
    return sum([j - i + 1 for (i, j) in joined_intervals])


def blast_results_filter(
    blast_tsv,
    fastx_file,
    out_file,
    terms,
    contaminated_contigs=None,
    length=1000,
    identity=95.0,
):
    """Filter FASTX based on intervals matching terms within BLAST results.

    Args:
        blast_tsv (str or list): tsv of BLAST results. See note
        fastx_file (str): file path to existing file
        out_file (str): file path to be written
        terms (str or list): search terms to be considered a hit, e.g. ["Homo sapiens genomic DNA"]
        contaminated_contigs (str): path to file to store contigs marked as contaminants
        length (Optional[int]): required match length to be considered a hit
        identity (Optional[float]): required matched identity to be filtered

    Returns:
        str

    Note:
        blast tsv requires a header with at least the following fields:
            Query Seq-id
            Percentage of identical matches
            Alignment length
            All Subject Title(s)
    """
    if file_exists(out_file):
        return out_file
    if file_exists(out_file + ".gz"):
        return out_file + ".gz"

    # _csv.Error: field larger than field limit (131072)
    import sys
    import csv

    max_int = sys.maxsize
    decrement = True

    while decrement:
        # decrease the max_int value by factor 10
        # as long as the OverflowError occurs.
        decrement = False
        try:
            csv.field_size_limit(max_int)
        except OverflowError:
            max_int = int(max_int / 10)
            decrement = True

    # start the filter
    if isinstance(blast_tsv, six.string_types):
        blast_tsv = [blast_tsv]
    for btsv in blast_tsv:
        if not file_exists(btsv):
            logger.critical("%s not found." % btsv)
            return ""
    if not file_exists(fastx_file):
        logger.critical("%s not found." % fastx_file)
        return ""

    logger.info("Filtering %s based on BLAST results." % fastx_file)
    if isinstance(terms, six.string_types):
        terms = [terms.lower()]
    else:
        terms = [t.lower() for t in terms]
    # pull out IDs of contaminated entries from blast results
    contaminated = set()
    for btsv in blast_tsv:
        for qseq_id, alignments in groupby(
            reader(btsv, header=True, sep="\t"), key=lambda x: x["Query Seq-id"]
        ):
            # get the coordinates of passing hits
            coords = set()
            for alignment in alignments:
                if float(
                    alignment["Percentage of identical matches"]
                ) > identity and any(
                    t in alignment["All Subject Title(s)"].lower() for t in terms
                ):
                    coords.add(
                        (
                            int(alignment["Start of alignment in query"]),
                            int(alignment["End of alignment in query"]),
                        )
                    )
            # sum potentially overlapping coordinates
            if coords and overlap_distance(list(coords)) > length:
                contaminated.add(qseq_id)

    # writes empty file when every contig is contaminated
    # filter the contig into out_file
    if contaminated_contigs:
        contaminated_contigs_fh = open(contaminated_contigs, "w")
    with file_transaction(out_file) as tx:
        with open(tx, "w") as fh:
            for name, seq, qual in readfx(fastx_file):
                if name in contaminated:
                    print_fasta_record(name, seq, contaminated_contigs_fh)
                else:
                    if qual is None:
                        print_fasta_record(name, seq, fh)
                    else:
                        print("@" + name, seq, "+", qual, sep="\n", file=fh)
    if contaminated_contigs:
        contaminated_contigs_fh.close()
    return out_file


def run_checkm(fasta, out_dir, cores):
    logger.info("Running checkm on %s and writing to %s" % (fasta, out_dir))

    prefix = op.basename(fasta).split(".")[0]
    fasta = op.abspath(fasta)
    completeness = 0.0
    number_unique_markers = 0
    number_multi_copy = 0
    taxonomy_contained = "NA"
    taxonomy_sister_lineage = "NA"

    # we want the fasta in a directory by itself
    with tmp_dir() as bin_path:
        tmp_fasta = op.join(bin_path, op.basename(fasta))
        shutil.copy(fasta, tmp_fasta)

        plot_path = op.join(out_dir, "plots")

        # lineage workflow
        if not file_exists(op.join(out_dir, "completeness.tsv")):
            # clear any existing output directory and create
            shutil.rmtree(out_dir, ignore_errors=True)
            safe_makedir(out_dir)
            # run lineage workflow
            logger.info("Running lineage workflow on %s" % fasta)
            run(
                (
                    "checkm lineage_wf -f {outdir}/completeness.tsv --tab_table "
                    "-q -x fasta -t {cores} {binpath} {outdir}"
                ).format(outdir=out_dir, cores=cores, binpath=bin_path)
            )
        try:
            for l in reader(out_dir + "/completeness.tsv", header=True, sep="\t"):
                completeness = l["Completeness"]
                break
        except IOError:
            logger.warning("Lineage workflow failed for %s" % fasta)
            pass

        # tree (has better output than workflow)
        if not file_exists(op.join(out_dir, "taxonomy.tsv")):
            logger.info("Getting taxonomy")
            # checkm tree -t 30 -x fasta --tmpdir /dev/shm fastas checkm_o
            # checkm tree_qa --tab_table -o 2 checkm_o > taxonomy.tsv
            # Bin Id	# unique markers (of 43)	# multi-copy	Insertion branch UID	Taxonomy (contained)	Taxonomy (sister lineage)	GC	Genome size (Mbp)	Gene counCoding density	Translation table	# descendant genomes	Lineage: GC mean	Lineage: GC std	Lineage: genome size (Mbp) mean	Lineage: genome size (Mbp) std	Lineage: gene count mean	Lineage: gene count std
            # AG-323-G20_contigs	1	0	UID2169	k__Bacteria;p__Cyanobacteria;c__Chroococcales;o__Chroococcales;f__Cyanobium	f__Synechococcus;g__Synechococcus;s__	43.97247688	0.398356	453	0.773782747091	11	2	53.2604384148	0.813097897058	2.656719	0.049971	2998.5	47.5
            run(
                (
                    "checkm tree_qa --tab_table -o 2 {outdir} "
                    "-f {outdir}/taxonomy.tsv"
                ).format(outdir=out_dir)
            )
        try:
            for l in reader("%s/taxonomy.tsv" % out_dir, header=True, sep="\t"):
                number_unique_markers = l["# unique markers (of 43)"]
                number_multi_copy = l["# multi-copy"]
                taxonomy_contained = l["Taxonomy (contained)"]
                taxonomy_sister_lineage = l["Taxonomy (sister lineage)"]
                break
        except IOError:
            logger.warning("Failed to retrieve taxonomy info")
            pass

        # try to make these plots:
        try:
            # bin_qa_plot
            if not file_exists(op.join(plot_path, "bin_qa_plot.pdf")):
                logger.info("Running bin_qa_plot on %s" % fasta)
                run(
                    (
                        "checkm bin_qa_plot -x fasta -q --image_type pdf {outdir} "
                        "{binpath} {plotpath}"
                    ).format(outdir=out_dir, binpath=bin_path, plotpath=plot_path)
                )
            # marker_plot
            if not file_exists(op.join(plot_path, prefix + ".marker_pos_plot.pdf")):
                logger.info("Running marker_plot on %s" % fasta)
                run(
                    (
                        "checkm marker_plot -x fasta -q --image_type pdf {outdir} "
                        "{binpath} {plotpath}"
                    ).format(outdir=out_dir, binpath=bin_path, plotpath=plot_path)
                )
            # dist_plot
            if not file_exists(op.join(plot_path, prefix + ".ref_dist_plots.pdf")):
                logger.info("Running dist_plot on %s" % fasta)
                tns = os.path.join(out_dir, "tetranucleotide_signatures.tsv")
                run(
                    "checkm tetra -q -t {cores} {fa} {tns}".format(
                        cores=cores, fa=fasta, tns=tns
                    )
                )
                run(
                    (
                        "checkm dist_plot -q -x fasta --image_type pdf {outdir} "
                        "{binpath} {plotpath} {tns} 95"
                    ).format(
                        outdir=out_dir, binpath=bin_path, plotpath=plot_path, tns=tns
                    )
                )
        # subprocess.CalledProcessError
        except:
            pass
    return (
        completeness,
        number_unique_markers,
        number_multi_copy,
        taxonomy_contained,
        taxonomy_sister_lineage,
    )


def classify_fasta(
    fasta,
    blastdb,
    rrna_file,
    min_score=150,
    num_alignments=10,
    min_rrna_length=75,
    threads=8,
):
    cmd = (
        "blastn -task megablast -query {q} -db {db} "
        "-num_alignments {na} -outfmt {fmt} -out {o} "
        "-num_threads {nt}"
    )
    with tmp_dir() as tmp:
        blast_hits = True
        xml = os.path.join(tmp, op.basename(fasta) + ".xml")
        run(
            cmd.format(
                q=fasta, db=blastdb, na=num_alignments, fmt=5, o=xml, nt=threads
            ),
            description="BLAST against %s" % blastdb,
        )
        crest_dir = os.path.join(tmp, op.basename(fasta) + ".crest")
        # should include a DB option for CREST call...
        try:
            run(
                "classify {xml} -o {cd} --minscore={ms}".format(
                    xml=xml, cd=crest_dir, ms=min_score
                )
            )
        except:
            blast_hits = False
            pass

        # skip if the other blast search yielded nothing; determined by classification failure
        if blast_hits:
            with open(rrna_file, "w") as rrna_fh:
                for qseq_id, alignments in groupby(
                    run(
                        cmd.format(
                            q=fasta,
                            db=blastdb,
                            na=num_alignments,
                            fmt='"6 std qseq"',
                            o="-",
                            nt=threads,
                        ),
                        iterable=True,
                    ),
                    key=lambda x: x.strip().split("\t")[0],
                ):
                    for alignment in alignments:
                        # take best hit only per contig; no gaps in seq
                        alignment = alignment.strip().split("\t")
                        if not int(alignment[3]) >= min_rrna_length:
                            continue
                        print_fasta_record(qseq_id, alignment[-1], rrna_fh)
                        break

        # Level	Taxonpath	Taxon	xxxAG-459-P16
        # base	root;No hits	No hits	30355
        # base	root;Cellular organisms	Cellular organisms	0
        # domain	root;Cellular organisms;Bacteria	Bacteria	0
        # phylum	root;Cellular organisms;Bacteria;Cyanobacteria	Cyanobacteria	0

        taxonpath = "root;No hits"
        count = 0
        if file_exists(op.join(crest_dir, "All_Assignments.tsv")):
            with open(op.join(crest_dir, "All_Assignments.tsv")) as fh:
                for line in fh:
                    toks = line.strip().split("\t")
                    if "root;No hits" in toks[1] or "Taxonpath" in toks[1]:
                        continue
                    if int(toks[3]) > count:
                        count = int(toks[3])
                        taxonpath = toks[1]
        return taxonpath


def run_seqtk_sample(fastqs, out_files, n, seed=37):
    """Subsample incoming paired-end fastqs to `n` reads (serially).

    Args:
        fastqs (list): list of fastq paths
        out_files (list): list of output fastq paths; output files are always gzipped
        n (int): number of subsampled reads
        seed (int): for random selection of reads

    Returns:
        list: subsampled r1, r2 file paths
    """
    if file_exists(out_files):
        return out_files

    logger.info("Subsampling to %d reads" % n)
    with file_transaction(out_files) as tx:
        cmd = "seqtk sample -s {seed} {fastq} {number} | gzip > {out}".format(
            seed=seed, fastq=fastqs[0], number=n, out=tx[0]
        )
        run(cmd)
        cmd = "seqtk sample -s {seed} {fastq} {number} | gzip > {out}".format(
            seed=seed, fastq=fastqs[1], number=n, out=tx[1]
        )
        run(cmd)
    return out_files


def run_flash(
    fastq, out_file, mismatch_density=0.05, min_overlap=35, max_overlap=150, cores=1
):
    """Joins interleaved FASTQ file using `flash`.

    Args:
        fastq (str): file path to fastq
        out_file (str): path of desired gzip compressed fastq
        mismatch_density (Optional[float]): mismatch density of overlapping region
        min_overlap (Optional[int]): minimum expected read overlap
        max_overlap (Optional[int]): maximum expected read overlap
        cores (Optional[int]): threads allocated to flash

    Returns:
        str
    """
    # in case we switch what the current config looks like
    # predefined_options = [('--interleaved-input', True),
    #                       ('--interleaved', True), ('-I', True),
    #                       ('-Ti', True), ('--tab-delimited-input', True),
    #                       ('-To', True), ('--tab-delimited-output', True),
    #                       ('-o', False), ('--output-prefix', False),
    #                       ('-d', False), ('--output-directory', False),
    #                       ('-c', True), ('--to-stdout', True),
    #                       ('-z', True), ('--compress', True),
    #                       ('--compress-prog', True),
    #                       ('--compress-prog-args', True), ('--suffix', True),
    #                       ('--output-suffix', True), ('-t', False),
    #                       ('--threads', False), ('-m', False),
    #                       ('--min-overlap', False), ('-M', False),
    #                       ('--max-overlap', False), ('-x', False),
    #                       ('--max-mismatch-density', False)]
    # options = filter_options(options, predefined_options)

    if file_exists(out_file):
        return out_file
    if not out_file.endswith(".gz"):
        out_file = out_file + ".gz"
        if file_exists(out_file):
            return out_file

    logger.info("Joining %s using flash" % fastq)

    with file_transaction(out_file) as tx:
        cmd = (
            "flash --interleaved-input -x {mismatch} -m {min} "
            "-M {max} -t {threads} --to-stdout {fastq} "
            "| gzip > {tx}"
        ).format(
            mismatch=mismatch_density,
            min=min_overlap,
            max=max_overlap,
            threads=cores,
            fastq=fastq,
            tx=tx,
        )
        run(cmd, description="Joining reads with flash")
    return out_file


def flash_version():
    """Get the version from `flash --version`.

    Returns:
        str
    """
    for line in run("flash --version", iterable=True):
        return line.strip()


def run_prokka(fasta, out_dir, prefix, proteins=None, protein_out=None, cores=1):
    """prefix is used for the file name and the locus tag

    outputs: .err .faa .ffn .fna .fsa .gbk .gff .log .sqn .tbl .tsv .txt
    """

    prokka_exts = [
        ".err",
        ".faa",
        ".ffn",
        ".fsa",
        ".gff",
        ".log",
        ".tbl",
        ".tsv",
        ".txt",
        ".fna",
    ]
    out_files = [op.join(out_dir, prefix + x) for x in prokka_exts]
    if proteins:
        if not protein_out:
            protein_out = op.join(out_dir, prefix + "_prokka-swissprot.tsv")
        out_files.append(protein_out)
    if file_exists(out_files):
        logger.info("Prokka is complete for %s" % prefix)
        return out_files

    # with file_transaction(out_files) as tx:
    cmd = (
        "prokka --force --outdir {out} --prefix {prefix} --locustag {prefix}"
        " --cpus {cores} {fasta}"
    ).format(out=out_dir, prefix=prefix, cores=cores, fasta=fasta)
    prokka_gff = op.join(out_dir, prefix + ".gff")
    try:
        run(cmd, description="Annotation with Prokka")
    except:
        logger.warning("Annotation with Prokka of %s failed" % prefix)
        pass

    if proteins:
        out_dir = op.dirname(op.abspath(protein_out))
        prot_files = [op.join(out_dir, prefix + "-proteins" + x) for x in prokka_exts]
        cmd = (
            "prokka --force --outdir {out} --prefix {prefix}-proteins --locustag {prefix}"
            " --cpus {cores} --rawproduct --proteins {proteins}"
            " {fasta}"
        ).format(
            out=out_dir, prefix=prefix, cores=cores, proteins=proteins, fasta=fasta
        )

        try:
            run(cmd, description="Secondary annotation with Prokka")

            gff = op.join(out_dir, prefix + "-proteins.gff")

            locus_tag_re = re.compile(r"locus_tag=(.*?)(?:;|$)")
            ec_re = re.compile(r"eC_number=(.*?)(?:;|$)")
            gene_re = re.compile(r"gene=(.*?)(?:;|$)")
            product_re = re.compile(r"product=(.*?)(?:;|$)")

            prokka_annotation = dict()
            # needs a gff generator
            with open(prokka_gff) as gff_fh:
                for line in gff_fh:
                    if line.startswith("##FASTA"):
                        break
                    if line.startswith("#"):
                        continue
                    toks = line.strip().split("\t")
                    if not toks[2] == "CDS":
                        continue
                    try:
                        locus_tag = locus_tag_re.findall(toks[-1])[0]
                    except IndexError:
                        locus_tag = ""
                    if not locus_tag:
                        logger.critical(
                            "Unable to locate a locus tag in [%s]" % toks[-1]
                        )
                        sys.exit(1)
                    try:
                        gene = gene_re.findall(toks[-1])[0]
                    except IndexError:
                        gene = ""
                    try:
                        ec_number = ec_re.findall(toks[-1])[0].replace("%2", ",")
                    except IndexError:
                        ec_number = ""
                    try:
                        product = product_re.findall(toks[-1])[0].replace("%2", ",")
                    except IndexError:
                        product = ""
                    prokka_annotation[toks[0] + "-" + toks[3] + "-" + toks[4]] = {
                        "contig_id": toks[0],
                        "gene": gene,
                        "EC_number": ec_number,
                        "product": product,
                    }

            with open(gff) as gff_fh, open(protein_out, "w") as output:
                # print the header into the output file
                print(
                    "contig_id",
                    "locus_tag",
                    "ftype",
                    "start",
                    "stop",
                    "strand",
                    "prokka_gene",
                    "prokka_EC_number",
                    "prokka_product",
                    "swissprot_gene",
                    "swissprot_EC_number",
                    "swissprot_product",
                    "swissprot_eggNOG",
                    "swissprot_KO",
                    "swissprot_Pfam",
                    "swissprot_CAZy",
                    "swissprot_TIGRFAMs",
                    sep="\t",
                    file=output,
                )

                for line in gff_fh:
                    if line.startswith("##FASTA"):
                        break
                    if line.startswith("#"):
                        continue
                    toks = line.strip().split("\t")
                    if not toks[2] == "CDS":
                        continue
                    try:
                        locus_tag = locus_tag_re.findall(toks[-1])[0]
                    except IndexError:
                        locus_tag = ""
                    if not locus_tag:
                        logger.critical(
                            "Unable to locate a locus tag in [%s]" % toks[-1]
                        )
                        sys.exit(1)
                    try:
                        gene = gene_re.findall(toks[-1])[0]
                    except IndexError:
                        gene = ""
                    try:
                        ec_number = ec_re.findall(toks[-1])[0].replace("%2", ",")
                    except IndexError:
                        ec_number = ""
                    try:
                        product = product_re.findall(toks[-1])[0]
                    except IndexError:
                        product = ""

                    product = product.replace("%2", ",")
                    additional_annotation = dict()
                    if "^^" in product:
                        for tokens in product.split("^^"):
                            items = tokens.split("::")
                            additional_annotation[items[0]] = items[1]
                    else:
                        additional_annotation = {
                            "product": product,
                            "eggNOG": "",
                            "KO": "",
                            "Pfam": "",
                            "CAZy": "",
                            "TIGRFAMs": "",
                        }

                    try:
                        prokka_entry = prokka_annotation[
                            toks[0] + "-" + toks[3] + "-" + toks[4]
                        ]
                    except KeyError:
                        # lines from the prokka annotation will be missing
                        logger.warning(
                            "Prokka annotation missing for %s-%s-%s"
                            % (toks[0], toks[3], toks[4])
                        )
                        prokka_entry = {"gene": "", "EC_number": "", "product": ""}

                    print(
                        toks[0],
                        locus_tag,
                        toks[2],
                        toks[3],
                        toks[4],
                        toks[6],
                        prokka_entry["gene"],
                        prokka_entry["EC_number"],
                        prokka_entry["product"],
                        gene,
                        ec_number,
                        additional_annotation["product"],
                        additional_annotation["eggNOG"],
                        additional_annotation["KO"],
                        additional_annotation["Pfam"],
                        additional_annotation["CAZy"],
                        additional_annotation["TIGRFAMs"],
                        sep="\t",
                        file=output,
                    )

            # delete all of the other files
            for f in prot_files:
                os.remove(f)
        except:
            logger.warning("Annotation with Prokka+Proteins of %s failed" % prefix)
            pass

    return out_files


def run_interproscan(fasta, out_file, tmp_dir=None, cores=1):
    if file_exists(out_file):
        return out_file

    if not tmp_dir:
        tmp_dir = tempfile.gettempdir()

    with file_transaction(out_file) as tx:
        cmd = (
            "interproscan.sh --tempdir {tmp_dir} "
            "--input {fasta} --formats tsv "
            "--pathways --iprlookup --goterms "
            "--cpu {cores} --outfile {out_file}"
        ).format(tmp_dir=tmp_dir, fasta=fasta, cores=cores, out_file=tx)
        try:
            run(cmd, description="Running InterProScan")
            if file_exists(tx):
                header = [
                    "Protein Accession",
                    "Sequence MD5 digest",
                    "Sequence Length",
                    "Analysis",
                    "Signature Accession",
                    "Signature Description",
                    "Start",
                    "Stop",
                    "Score",
                    "Status",
                    "Date",
                    "InterPro accession",
                    "InterPro description",
                    "GO annotations",
                    "Pathways annotations",
                ]
                for line in fileinput.input([tx], inplace=True):
                    if fileinput.isfirstline():
                        print(*header, sep="\t")
                    print(line.strip())
        except:
            logger.warning("Annotation with InterProScan on %s failed" % fasta)
            return False

    return out_file


def parse_prokka_output(path):
    data = dict(count="NA", hypothetical_fraction="NA", average_length="NA")
    if os.path.exists(path):
        with open(path) as fh:
            reader = csv.DictReader(fh, delimiter="\t")
            total = 0.0
            hypothetical = 0.0
            lengths = []
            for row in reader:
                total += 1
                if (
                    row["prokka_product"] == "hypothetical protein"
                    and row["swissprot_product"] == "hypothetical protein"
                ):
                    hypothetical += 1
                lengths.append(int(row["stop"]) - int(row["start"]))
            data["count"] = total
            data["hypothetical_fraction"] = hypothetical / total
            data["average_length"] = sum(lengths) / float(len(lengths))
    return data


def run_bbmerge(pe_fq, out_file, k=40, extend2=60, iterations=5, loose="t", qtrim2="t"):

    if file_exists(out_file):
        return out_file
    if not out_file.endswith(".gz"):
        out_file = out_file + ".gz"
        if file_exists(out_file):
            return out_file

    with file_transaction(out_file) as tx:
        cmd = (
            "bbmerge.sh -Xmx40G k={k} extend2={extend2} "
            "iterations={iterations} loose={loose} qtrim2={qtrim2} "
            "in={pe_fq} out={tx}"
        ).format(
            k=k,
            extend2=extend2,
            iterations=iterations,
            loose=loose,
            qtrim2=qtrim2,
            pe_fq=pe_fq,
            tx=tx,
        )
        run(cmd, description="Joining PE reads with bbmerge")
    return out_file



from __future__ import print_function
import logging
import multiprocessing
import operator
import os
import sys
from itertools import groupby
from pysam import FastxFile
from toolshed import nopen

try:
    from itertools import izip as zip
except ImportError:
    pass

from .utils import file_exists, file_transaction, multiprocess, cp, grouper, run, pigz_file


logger = logging.getLogger(__name__)


def readfx(fastx):
    if not file_exists(fastx):
        logger.critical("File Not Found: %s" % fastx)
        raise IOError(2, "No such file:", fastx)

    fx = ""
    try:
        fx = FastxFile(fastx)
        for f in fx:
            yield f.name, f.sequence, f.quality
    finally:
        if fx:
            fx.close()


def read_fasta(fh):
    """FASTA iterator where name isn't split by space."""
    for header, group in groupby(fh, lambda line: line[0] == '>'):
        if header:
            line = next(group)
            name = line[1:].strip()
        else:
            seq = ''.join(line.strip() for line in group)
            yield name, seq


def complexity_filter_sequence(seq, threshold=0.05, alphabet='ACGT'):
    """
    filter sequence based on a given expectation of bases above an observed
    threshold.

    parameters
        seq : sequence string
        threshold : minimal observed composition for each character in alphabet
        alphabet : verified sequence characters above threshold

    returns
        boolean
    """
    count_threshold = len(seq) * threshold
    for elem in alphabet:
        if seq.count(elem) < count_threshold:
            return False
    return True


def complexity_filter_pe_record(complextuple, threshold, alphabet):
    """
    mapped function to filter 2 records at once.

    parameters
        n1, s1, q1 : tuple of first record's name, sequence, and quality values
        n2, s2, q2 : tuple of second record
        threshold : minimal observed fraction for each character in alphabet
        alphabet : verified sequence characters above threshold

    returns
        full records string on success. empty string on fail.
    """
    ((n1, s1, q1), (n2, s2, q2)) = complextuple
    try:
        assert n1 == n2
    except AssertionError:
        assert n1[:-2] == n2[:-2], "Pairing failed due to sync: %s %s" % (n1, n2)

    pass1 = complexity_filter_sequence(s1, threshold, alphabet)
    pass2 = complexity_filter_sequence(s2, threshold, alphabet)

    if pass1 and pass2:
        if not n1.endswith('/1'): n1 += '/1'
        if not n2.endswith('/2'): n2 += '/2'
        records = "@%s\n%s\n+\n%s\n@%s\n%s\n+\n%s" % (n1, s1, q1, n2, s2, q2)
        return records
    return ""


def complexity_filter(r1, r2, out_file, threshold=0.05, cores=1):
    """
    filters fastq based on read composition above threshold.

    parameters
        r1 : R1 fastq
        r2 : R2 fastq
        out_file : interleaved fastq file path
        threshold : fraction of required bases (ACTG) per read sequence
        cores : number of cores

    returns
        a tuple (out_file: string, surviving read pairs: int)
    """
    if file_exists(out_file):
        return out_file, read_count(out_file) / 2

    gzip = False
    if out_file.endswith(".gz"):
        gzip = True
        out_file = out_file.rpartition('.gz')[0]

    count_file = out_file + '.count'
    total_pairs = 0
    filtered_pairs = 0
    p = multiprocessing.Pool(cores)

    logger.info("Complexity filtering. Threshold is %0.02f" % threshold)

    with file_transaction(out_file) as tx_out_file:
        with open(tx_out_file, 'w') as out_handle:
            for results in multiprocess(complexity_filter_pe_record,
                    zip(readfx(r1), readfx(r2)), threshold, 'ACGT', pool=p):
                for record in results:
                    total_pairs += 1
                    if not record:
                        filtered_pairs += 1
                        continue
                    print(record, file=out_handle)
    surviving_pairs = total_pairs - filtered_pairs
    if gzip:
        # returns with .gz added to out_file
        out_file = pigz_file(out_file, cores)

    logger.info("Filtering complete. Observed pairs: %d. Total surviving: %d" % (total_pairs, surviving_pairs))
    with open(count_file, 'w') as out_handle:
        print(surviving_pairs * 2, file=out_handle)

    return out_file, surviving_pairs


def read_count(fname):
    """Count the number of reads and write metadata .count file.

    Args:
        fname (str): fastq or fasta file path

    Returns:
        int
    """
    if not file_exists(fname):
        return 0
    total = 0
    fq = True
    for name, seq, qual in readfx(fname):
        if not qual:
            fq = False
        break

    if fname.endswith("gz"):
        count_file = fname.rsplit(".gz", 1)[0] + ".count"
        cat = "gzip -d -c"
    else:
        count_file = fname + ".count"
        cat = "cat"

    if file_exists(count_file):
        with open(count_file) as fh:
            for line in fh:
                total = int(line.strip())
                return total

    if not fq:
        cmd = '%s %s | grep -c "^>"' % (cat, fname)
    else:
        cmd = '%s %s | wc -l' % (cat, fname)

    for line in run(cmd, description="Counting reads", iterable=True):
        total = int(line.rstrip())
        if fq:
            assert total % 4 == 0, "Multi-line or invalid FASTQ"
            total = int(total / 4)

    with open(count_file, 'w') as fh:
        print(total, file=fh)

    return total


def print_fasta_record(name, seq, out_handle=sys.stdout, wrap=80):
    """
    print fasta record and wraps the sequence line.

    parameters
        name : name or header for fasta entry
        seq : sequence
        out_handle : open file handle in which to write or stdout
        wrap : line width of fasta sequence
    """
    print('>', name, sep='', file=out_handle)
    if wrap:
        for i in range(0, len(seq), wrap):
            print(seq[i:i + wrap], file=out_handle)
    else:
        print(seq, file=out_handle)


def format_fasta_record(name, seq, wrap=100):
    """Fasta __str__ method.

    Convert fasta name and sequence into wrapped fasta format.

    Args:
        name (str): name of the record
        seq (str): sequence of the record
        wrap (int): length of sequence per line

    Yields:
        tuple: name, sequence

    >>> format_fasta_record("seq1", "ACTG")
    ">seq1\nACTG"
    """
    record = ">" + name + "\n"
    if wrap:
        for i in range(0, len(seq), wrap):
            record += seq[i:i+wrap] + "\n"
    else:
        record += seq + "\n"
    return record.strip()


def munge_header(fastx, out_file, text, intent='prepend', sep='_'):
    """
    Prepend or append text to name field of fastx using sep.

    parameters
        fastx : fasta or fastq
        out_file : result file
        text : string to append to name field
        intent : 'prepend' or 'append' action
        sep : separator to use between text and existing name

    returns
        output file name : string
    """
    if file_exists(out_file):
        return out_file

    with file_transaction(out_file) as tx_out_file:
        with open(tx_out_file, 'w') as out_handle:
            for name, seq, qual in readfx(fastx):
                if intent == 'prepend':
                    name = "%s%s%s" % (text, sep, name)
                elif intent == 'prepend-and-split':
                    # update the fasta headers from
                    # AG-325-P19_NODE_1_length_12099_cov_58.823
                    # to
                    # AG-325-P19_NODE_1
                    name = name.partition('_length')[0]
                    name = "%s%s%s" % (text, sep, name)
                else:
                    name = "%s%s%s" % (name, sep, text)

                # fastq
                if qual:
                    print("@" + name, seq, "+", qual, sep="\n", file=out_handle)
                # fasta
                else:
                    print_fasta_record(name, seq, out_handle)
    return out_file


def length_filter(fasta, pass_file, fail_file, operation='gt', length=2000, wrap=60):
    """
    filter fasta sequences by length. pass_file meets criteria of operation.

    parameters
        fasta : fasta sequence file
        pass_file : sequences that pass criteria
        fail_file : sequences that do not pass criteria
        operation : 'gt' or 'lt'; keep sequences of length greater than or
            less than respectively
        length : length threshold
        wrap : line width of fasta sequence

    returns
        pass_file and fail_file : tuple
    """
    if file_exists([pass_file, fail_file]):
        return pass_file, fail_file

    logger.info('Filtering %s by length %d' % (fasta, length))

    op_func = operator.gt if operation == 'gt' else operator.lt
    with file_transaction([pass_file, fail_file]) as tx_out_files:
        with open(tx_out_files[0], 'w') as passfh, open(tx_out_files[1], 'w') as failfh:
            for name, seq, qual in readfx(fasta):
                if op_func(len(seq), length):
                    print_fasta_record(name, seq, passfh, wrap)
                else:
                    print_fasta_record(name, seq, failfh, wrap)
    return pass_file, fail_file


def bedtools_genome(fasta, out_file):
    with open(out_file, 'w') as fh:
        for name, seq, qual in readfx(fasta):
            print(name, len(seq), sep="\t", file=fh)


def fasta_stats(fasta):
    """
    calculate basic stats from a fasta file.

    parameters
        fasta : fasta file path

    returns
        total bases, max contig size, contig count, gc percent : tuple
    """
    logger.info('Calculating stats for %s' % fasta)

    if read_count(fasta) == 0:
        return 0, 0, 0, 0

    total_contigs = 0
    contig_sizes = []
    gc_total = 0

    for i, (name, seq, qual) in enumerate(readfx(fasta), start=1):
        contig_sizes.append(len(seq))
        gc_total += seq.count('G') + seq.count('C')
        total_contigs = i

    if total_contigs > 0:
        contig_sizes.sort(reverse=True)
        max_contig_size = contig_sizes[0]
        total_bases = sum(contig_sizes)
        gc_percent = float(gc_total) / total_bases * 100

    else:
        total_bases = 0
        max_contig_size = 0
        gc_percent = 0

    return total_bases, max_contig_size, total_contigs, gc_percent


def length_sort(fasta, out_file):
    """
    sort a fasta file by sequence length.

    parameters
        fasta : fasta file path
        out_file : output file path

    returns
        output file path : string
    """
    if file_exists(out_file):
        return out_file

    logger.info('Sorting %s sequences by length' % fasta)

    records = [(name, seq) for name, seq, qual in readfx(fasta)]
    records.sort(key=lambda x: len(x[1]), reverse=True)

    with file_transaction(out_file) as tx_out_file:
        with open(tx_out_file, 'w') as out_handle:
            for name, seq in records:
                print_fasta_record(name, seq, out_handle)
    return out_file


def remove_seq_wraps(fasta, out_file):
    """
    Remove line wraps in the sequence string of a fasta.

    parameters
        fasta : fasta file path
        out_file : output file

    returns
         output file path : string
    """
    if file_exists(out_file):
        return out_file

    logger.info("Removing line wraps in %s" % fasta)

    with file_transaction(out_file) as tx_out_file:
        with open(tx_out_file, 'w') as out_handle:
            for name, seq, qual in readfx(fasta):
                print(">%s\n%s" % (name, seq), file=out_handle)

    return out_file


def calc_skew_and_content(seq, window_size=500):
    """
    sliding window implementation of gc_skew and gc_content.

    parameters
        seq : sequence string to analyze
        window_size : length of string to analyze at a time

    returns
        mid point location of window, skew, gc content : tuple

    """
    half_window = window_size / 2
    seq = seq.upper()
    seq_len = len(seq)
    if not seq_len >= window_size:
        logger.warn("Sequence too short: %s" % seq)
        yield None, None, None

    # could likely be improved by using numpy.chararray
    for start in range(0, seq_len - window_size + 1):
        stop = start + window_size
        mid = stop - half_window
        s = seq[start:stop]

        g = s.count('G')
        c = s.count('C')
        gc = g + c

        content = float(gc) / window_size
        skew = 0 if gc == 0 else (g - c) / float(g + c)
        yield mid, skew, content


def gc_skew_and_content(fasta, out_file, window_size=500):
    """
    Calculate GC content and skew for a fasta file over a sliding window of
    each sequence.

    parameters
        fasta : fasta file path
        out_file : output file path
        window_size : sliding window size on which to analyze sequence

    returns
        output file path : string
    """
    if file_exists(out_file):
        return out_file

    header = ["SEQUENCE_NAME", "POSITION", "SKEW", "CONTENT"]
    with file_transaction(out_file) as tx_out_file:
        with open(tx_out_file, 'w') as out_handle:
            print(*header, sep="\t", file=out_handle)
            for name, seq, qual in readfx(fasta):
                for point, skew, content in calc_skew_and_content(seq, window_size):
                    if not point:
                        continue
                    print(name, point, skew, content, sep="\t", file=out_handle)

    return out_file


def trim(s, left, right):
    """
    trim('AAAACCCCCAAAA', 4, 4)
    'CCCCC'
    trim('AAAACCCCCAAAA', 4, 0)
    'CCCCCAAAA'
    trim('AAAACCCCCAAAA', 0, 4)
    'AAAACCCCC'
    """
    return s[left:][:-right] if right > 0 else s[left:]


def trim_fastx(fastx_file, out_file, left=0, right=0):
    """
    Trim sequence (and qual) based on left and right options.

    fastx_file : fasta or fastq sequence file path
    out_file : output file path
    left : number of bases to trim from 5' end
    right : number of bases to trim from 3' end

    returns out_file path : string
    """
    assert left >= 0 and right >= 0

    if file_exists(out_file):
        return out_file

    if left == 0 and right == 0:
        cp(fastx_file, out_file)
        return out_file

    logger.info("Trimming the contigs of %s" % fastx_file)

    with file_transaction(out_file) as tx_out_file:
        with open(tx_out_file, 'w') as out_handle:
            for name, seq, qual in readfx(fastx_file):
                seq = trim(seq, left, right)
                if not seq: continue
                if qual:
                    qual = trim(qual, left, right)
                    print('@' + name, seq, '+', qual, sep='\n', file=out_handle)
                else:
                    print_fasta_record(name, seq, out_handle)

    return out_file


def check_sync(name1, name2):
    try:
        assert name1 == name2
    except AssertionError:
        assert name1[:-2] == name2[:-2], \
            "Run failed due to read sync: %s %s" % (name1, name2)


def tmp_split_reads(reads, tmp="/tmp"):
    """
    split interleaved fastq into R1 and R2 in the context of not having to save
    the two fastqs for later use.
    """
    r1 = os.path.join(tmp, "r1.fastq")
    r2 = os.path.join(tmp, "r2.fastq")
    with open(r1, 'w') as fh1, open(r2, 'w') as fh2:
        for (n1, s1, q1), (n2, s2, q2) in grouper(2, readfx(reads)):
            check_sync(n1, n2)
            print('@' + n1, s1, '+', q1, sep='\n', file=fh1)
            print('@' + n2, s2, '+', q2, sep='\n', file=fh2)
    return r1, r2



from __future__ import print_function
import csv
import logging
import multiprocessing
import os
import os.path as op
import shutil
import six
import sys
import tarfile

from collections import Counter, OrderedDict
from glob import glob

import click
import click_config
import pandas
import parmap

from genologics.lims import Lims
from scgc import apps, classifier, fastx, utils


class config(object):
    """
    Define default values that may be missing in the configuration file.
    """

    class shared(object):
        data = ""
        results = ""
        deliverables = ""
        cores = 10
        min_score = 150
        top_fraction = 0.9
        min_length = 100
        cores = 10
        baseuri = "https://scgc-clarity.bigelow.org"
        username = "apiuser"
        password = "meantdiagramcreateprize"
        version = "v2"
        layouts_dir = "/mnt/scgc_raw/clarity/static/layout"

    class assemble(object):
        logs = "/mnt/scgc_raw/results/nextseq/logs/"
        bcl_queue = "scgc-route"
        queue = "route"
        time = "20:00:00"
        qsub_cpus = 8
        qsub_mem = "50G"
        job_name = "assembly"
        subsample = 0
        cores = 10
        skip_qc = False
        ignore_low_coverage = False
        trim_opts = "-phred33 LEADING:0 TRAILING:5 SLIDINGWINDOW:4:15 MINLEN:36"
        complexity_threshold = 0.05
        norm_opts = "-k 21 -t 30 -c 3"
        reference = ""
        reference_threshold = 0.05
        assembly_opts = "--careful --sc --phred-offset 33"
        passing_length = 2200
        left_trim = 100
        right_trim = 100
        blast_db = ""
        blast_opts = "-num_alignments 100"
        contamination_terms = "homo sapiens,human dna sequence"
        contamination_length = 100
        contamination_identity = 95.0
        pca_script = ""
        pca_opts = "--window 1600 --step 200"
        ssu_min_length = 500
        # SSU reference DB
        ssu_blast_db = ""
        ssu_blast_map = ""
        ssu_blast_tre = ""
        prokka_proteins = ""

    class classify(object):
        blast_db = ""
        function_map = ""

    class join(object):
        # could move some to shared space
        cores = 10
        skip_qc = False
        trim_opts = "-phred33 LEADING:0 TRAILING:5 SLIDINGWINDOW:4:15 MINLEN:75"
        complexity_threshold = 0.05
        mismatch_density = 0.05
        min_overlap = 35
        max_overlap = 150
        blast_opts = ""
        contamination_terms = ""
        contamination_length = 75
        contamination_identity = 95.0

    class deliverables(object):
        cores = 10

    class coassemble(object):
        outdir = "./results"


@click.group(context_settings=dict(help_option_names=["-h", "--help"]))
@click.version_option("0.4.3")
@click.pass_context
def cli(obj):
    """SCGC sequencing packages."""
    pass


def print_stats(out_file, stats_dict):
    # TODO "notes" placed ahead of checkM 'checkM_classification'
    with open(out_file, "w") as fh:
        print(*list(stats_dict.keys()), sep=",", file=fh)
        print(*list(stats_dict.values()), sep=",", file=fh)


def stop_processing(msg, stats=None, log_file=None, exit_code=0):
    logging.error(msg)
    if stats:
        stats["notes"] = msg
    if log_file:
        print_stats(log_file, stats)
    sys.exit(exit_code)


def process_samplesheet(
    samplesheet, new_samplesheet, reverse_complement=False, determine=False
):
    """Fix hidden characters in sample names and optional reverse complement
    the second index.

    Args:
        samplesheet (str): file path to SampleSheet.csv
        reverse_complement (bool): to reverse complement 'Index2'

    Returns:
        list
    """
    if not determine:
        logging.info("Processing %s", samplesheet)

    samples = []
    start = False
    index2_idx = None

    with open(samplesheet, "rU" if six.PY2 else "r") as ifh, open(
        new_samplesheet, "w"
    ) as ofh:
        for line in ifh:
            toks = line.strip().split(",")
            if not start:
                # table header processing
                if toks[0] == "Sample_ID":
                    start = True
                    if reverse_complement:
                        if "index2" in toks:
                            index2_idx = toks.index("index2")
                        elif "Index2" in toks:
                            index2_idx = toks.index("Index2")
                        else:
                            logging.warn("There is no Index2 to reverse complement")

            elif toks[0]:
                # convert underscores to dashes
                toks[0] = toks[0].replace("_", "-").replace(".", "-")
                toks[1] = toks[0]
                samples.append(toks[0])

                # only adjust on known index
                if reverse_complement and index2_idx:
                    toks[index2_idx] = complement(toks[index2_idx])[::-1]

            # remove blank lines at end of table
            else:
                break
            print(*[t.strip() for t in toks], sep=",", file=ofh)

    run = os.path.basename(os.path.dirname(samplesheet))
    if not determine:
        logging.info("Found %d samples for run %s", len(samples), run)
    return samples


@cli.command("annotate", short_help="run prokka with optional references")
@click.argument("fasta", type=click.Path(exists=True))
@click.argument("outdir")
@click.argument("prefix")
@click.option("--cores", help="number of cores used by prokka")
@click_config.wrap(module=config, sections=["assemble", "shared"])
def cli_run_prokka(fasta, outdir, prefix, cores):
    """Annotate FASTA and write files to OUTDIR where output files begin with
    PREFIX. Additional annotations from PROTEINS is written to
    PREFIX-extra.tsv
    """
    log = utils.logger(log_name="Annotate")
    log.info("Annotation started on %s" % fasta)
    utils.safe_makedir(outdir)
    log.info("Output is being written to %s/%s" % (outdir, prefix))
    if cores:
        config.assemble.cores = cores
    apps.run_prokka(
        fasta,
        outdir,
        prefix,
        proteins=config.assemble.prokka_proteins,
        cores=config.assemble.cores,
    )
    log.info("Annotation complete.")


@cli.command("assemble", short_help="paired-end assembly pipeline")
@click.argument("r1")
@click.argument("run_id")
@click.option("--data", help="directory of raw sequence data; the parent of run data")
@click.option(
    "--results",
    help="parent directory for all result files; the parent of run result data",
)
@click.option(
    "--output",
    help="output path to overwrite configuration settings for results location",
)
@click.option("--cores", type=int, help="number of multithreading tasks where possible")
@click.option(
    "--ignore-low-coverage",
    is_flag=True,
    help="set to allow samples with fewer than 1k reads to be processed",
)
@click.option(
    "--subsample",
    type=int,
    default=0,
    help="subsample original fastq prior to executing pipeline; 0 disables",
)
@click.option("--trim-opts", help="trimmomatic options")
@click.option(
    "--complexity-threshold", type=float, help="sequence complexity threshold"
)
@click.option("--norm-opts", help="kmernorm options")
@click.option(
    "--contamination-reference", help="contamination reference fasta file path"
)
@click.option(
    "--reference-threshold",
    type=float,
    help="allowable mismatch fraction to contamination reference",
)
@click.option("--assembly-opts", help="spades options")
@click.option("--passing-length", type=int, help="passing contig length")
@click.option("--left-trim", type=int, help="bases to trim from 5' end")
@click.option("--right-trim", type=int, help="bases to trim from 3' end")
@click.option(
    "--blast-db",
    help="blast database; multiple can be specified as comma separated list of paths",
)
@click.option("--blast-opts", help="blast options")
@click.option("--contamination-terms", help="blast contamination filter terms")
@click.option(
    "--contamination-length",
    type=int,
    help="minimum length of a hit to be considered; overlapping hits are merged",
)
@click.option(
    "--contamination-identity",
    type=float,
    help="minimum percentage of identical matches across alignment to be considered",
)
@click.option("--pca-script", help="tetramerPCA script path")
@click.option("--pca-opts", help="tetramerPCA options")
@click.option(
    "--ssu-min-length",
    type=int,
    default=500,
    help="SSU minimum sequence length of reported assignments",
)
@click.option("--ssu-blast-db", help="SSU annotation DB")
@click.option(
    "--ssu-blast-map", help="organism map with ID corresponding to ssu-blast-db"
)
@click.option(
    "--ssu-blast-tre", help="newick tree of ID corresponding to ssu-blast-map"
)
@click.option(
    "--prokka-proteins",
    help="fasta file of priority protein sequences for prokka annotation",
)
@click.option(
    "--seqtk-seed", default=11, show_default=True, help="seqtk random sample seed"
)
@click.option("--with-interproscan", is_flag=True, help="run interproscan across genes")
@click_config.wrap(module=config, sections=["assemble", "shared"])
def scgc_assemble(
    r1,
    run_id,
    data,
    results,
    output,
    cores,
    ignore_low_coverage,
    subsample,
    trim_opts,
    complexity_threshold,
    norm_opts,
    contamination_reference,
    reference_threshold,
    assembly_opts,
    passing_length,
    left_trim,
    right_trim,
    blast_db,
    blast_opts,
    contamination_terms,
    contamination_length,
    contamination_identity,
    pca_script,
    pca_opts,
    ssu_min_length,
    ssu_blast_db,
    ssu_blast_map,
    ssu_blast_tre,
    prokka_proteins,
    seqtk_seed,
    with_interproscan,
):
    """Assembly protocol. Any options specified on the command line overwrite
    defaults and any configuration file.
    """
    assert (
        "_R1"
    ) in r1, (
        "We're operating under the assumption of Illumina's file naming convention."
    )
    r2 = r1.replace("_R1", "_R2")
    sample = op.basename(r1).partition("_")[0]
    log = utils.logger(log_name="Assemble")
    log.info("Assembly started on %s" % sample)
    # we're going to rename "sample" to "sample-<subsample number>"
    try:
        subsample_n = int(max([config.assemble.subsample, subsample]))
    except ValueError:
        log.exception("Subsampling parameter should be an integer")
        raise

    # rename sample to include the subsample number if requested
    if subsample_n:
        sample = "%s-%d" % (sample, subsample_n)
    if data:
        config.shared.data = data
    if results:
        config.shared.results = results
    # create proper output directory structure
    if not output:
        output_dir = op.join(config.shared.results, run_id, sample)
    else:
        output_dir = op.abspath(output)
    if cores:
        config.assemble.cores = cores
    if ignore_low_coverage:
        config.assemble.ignore_low_coverage = True
    if trim_opts:
        config.assemble.trim_opts = trim_opts
    if complexity_threshold:
        config.assemble.complexity_threshold = complexity_threshold
    if norm_opts:
        config.assemble.norm_opts = norm_opts
    if contamination_reference:
        config.assemble.reference = contamination_reference
    if reference_threshold:
        config.assemble.reference_threshold = reference_threshold
    if assembly_opts:
        config.assemble.assembly_opts = assembly_opts
    if passing_length:
        config.assemble.passing_length = passing_length
    if left_trim:
        config.assemble.left_trim = left_trim
    if right_trim:
        config.assemble.right_trim = right_trim
    if blast_db:
        config.assemble.blast_db = blast_db
    if blast_opts:
        config.assemble.blast_opts = blast_opts
    if contamination_terms:
        config.assemble.contamination_terms = contamination_terms
    if contamination_length:
        config.assemble.contamination_length = contamination_length
    if contamination_identity:
        config.assemble.contamination_identity = contamination_identity
    if pca_script:
        config.assemble.pca_script = pca_script
    if pca_opts:
        config.assemble.pca_opts = pca_opts
    if ssu_min_length:
        config.assemble.ssu_min_length = ssu_min_length
    if ssu_blast_db:
        config.assemble.ssu_blast_db = ssu_blast_db
    if ssu_blast_map:
        config.assemble.ssu_blast_map = ssu_blast_map
    if ssu_blast_tre:
        config.assemble.ssu_blast_tre = ssu_blast_tre
    if prokka_proteins:
        config.assemble.prokka_proteins = prokka_proteins
    utils.safe_makedir(output_dir)
    log.info("Writing results to %s" % output_dir)
    # the final stats file we're writing out upon completion
    run_stats_file = op.join(output_dir, "%s_assembly_info.csv" % sample)
    run_stats = OrderedDict({"well": sample})

    sample_name_parts = sample.split("-")
    plate = "-".join(sample_name_parts[0:2])
    try:
        # sample well will fail with their "custom" samples
        sample_well = sample_name_parts[2]
        lims = Lims(
            config.shared.baseuri,
            config.shared.username,
            config.shared.password,
            config.shared.version,
        )

        # lets just fail for now
        # cp contents
        artifact_id = lims.get_containers(plate)[0].placements["A:1"].id
        mda_process = lims.get_processes(
            type="rtMDA-1", inputartifactlimsid=artifact_id
        )[0]
        cp_data = dict()
        for artifact in mda_process.shared_result_files():
            if artifact.name.endswith(".csv"):
                # can raise requests.exceptions.HTTPError
                try:
                    contents = lims.get_file_contents(artifact.files[0].id).split("\n")
                    plate_id = contents[0].split(" ")[1]
                    sample_ids = contents[contents.index("[Fit Parameters]") + 1].split(
                        ","
                    )[1:]
                    cp_values = contents[contents.index("[Fit Parameters]") + 3].split(
                        ","
                    )[1:]
                    cp_data = {
                        "{}-{}".format(plate_id, k): v
                        for k, v in six.moves.zip(sample_ids, cp_values)
                    }
                except Exception:
                    cp_data[sample] = "error connecting to LIMS"
                break

        # facs contents
        facs_process = lims.get_processes(type="FACS", inputartifactlimsid=artifact_id)[
            0
        ]
        layout_file = op.join(
            config.shared.layouts_dir,
            op.basename(facs_process.udf["Plate Layout File"]),
        )
        plate_contents = dict()
        with open(layout_file) as fh:
            lines = fh.readlines()
            if len(lines) == 1:
                lines = lines[0].split("\r")

            for y_index, line in enumerate(lines):
                if y_index == 0:
                    continue
                toks = line.strip().split("\t")
                row = toks[0]
                for x_index, row_value in enumerate(toks[1:], start=1):
                    well = "{row}{number}".format(row=row, number="%02d" % (x_index,))
                    plate_contents[well] = row_value
        run_stats["well_type"] = plate_contents[sample_well]
        run_stats["wga_cp"] = cp_data[sample]
    except:
        pass
    run_stats["wgs_run_id"] = run_id

    # log from the run's conversion from .bcl to .fastq
    # bcl2fastq_log = op.join(config.shared.data, run_id, "bcl2fastq.log")
    # if op.exists(bcl2fastq_log):
    #     run_stats['bcl2fastq_version'], run_stats[
    #         'bcl2fastq_parameters'
    #     ] = apps.parse_bcl2fastq_log(
    #         bcl2fastq_log
    #     )

    if not op.exists(r2):
        stop_processing("zero starting reads", run_stats, run_stats_file)

    run_stats["raw_read_count"] = fastx.read_count(r1)
    current_read_count = run_stats["raw_read_count"]
    # stop operating on samples that had very low read counts
    if current_read_count < 1000 and not config.assemble.ignore_low_coverage:
        stop_processing("too few starting reads", run_stats, run_stats_file)
    # create sub folders
    qc_dir = utils.safe_makedir(op.join(output_dir, sample + "_WGS_QC"))
    ssu_dir = utils.safe_makedir(op.join(qc_dir, "SSU_rRNA_recovery"))
    reads_dir = utils.safe_makedir(op.join(output_dir, sample + "_WGS_reads"))
    blast_dir = op.join(qc_dir, sample + "_all_contigs_blastn")
    checkm_dir = op.join(qc_dir, sample + "_final_contigs_checkm")
    prokka_dir = op.join(output_dir, sample + "_functional_annotation", "Prokka")
    if with_interproscan:
        interproscan_dir = utils.safe_makedir(
            op.join(output_dir, sample + "_functional_annotation", "InterProScan")
        )
    swissprot_dir = utils.safe_makedir(
        op.join(output_dir, sample + "_functional_annotation", "Swiss-Prot")
    )
    pca_dir = op.join(qc_dir, sample + "_final_contigs_tetramer_pca")
    # subsample
    if subsample_n:
        r1, r2 = apps.run_seqtk_sample(
            [r1, r2],
            [
                op.join(reads_dir, sample + "_R1.fastq.gz"),
                op.join(reads_dir, sample + "_R2.fastq.gz"),
            ],
            subsample_n,
            seed=seqtk_seed,
        )
        # run_stats['subsampled_to'] = subsample_n
        current_read_count = fastx.read_count(r1)
    # fastqc
    if not config.assemble.skip_qc:
        fastqc = apps.run_fastqc(
            [r1, r2], op.join(qc_dir, sample + "_fastqc"), config.assemble.cores
        )
    # trim
    if config.assemble.trim_opts:
        out_files = apps.trimmomatic(
            [r1, r2],
            [
                op.join(reads_dir, sample + "_trimmed_R1.fastq.gz"),
                op.join(reads_dir, sample + "_orphan_R1.fastq.gz"),
                op.join(reads_dir, sample + "_trimmed_R2.fastq.gz"),
                op.join(reads_dir, sample + "_orphan_R2.fastq.gz"),
            ],
            config.assemble.trim_opts,
            config.assemble.cores,
        )
        # run_stats['trimmer'] = apps.trimmomatic_version()
        # run_stats['trim_parameters'] = '"%s"' % config.assemble.trim_opts
        # run_stats['trimmed_read_count'] = fastx.read_count(out_files[0])
        # current_read_count = run_stats['trimmed_read_count']
        next_r1 = out_files[0]
        next_r2 = out_files[2]
    else:
        next_r1 = r1
        next_r2 = r2
    if not current_read_count:
        stop_processing("no reads survived trimming", run_stats, run_stats_file)
    # complexity filter
    pe_reads, complexity_read_count = fastx.complexity_filter(
        next_r1,
        next_r2,
        op.join(reads_dir, sample + "_pe.fastq.gz"),
        config.assemble.complexity_threshold,
        config.assemble.cores,
    )
    # run_stats['complexity_filter_threshold'] = "%0.02f%%" % (
    #     config.assemble.complexity_threshold * 100
    # )
    # run_stats['complexity_filtered_read_count'] = complexity_read_count
    current_read_count = complexity_read_count
    if not current_read_count:
        stop_processing("no reads after complexity filter", run_stats, run_stats_file)
    # digital normalization
    if config.assemble.norm_opts and current_read_count > 0:
        normed_reads = apps.run_kmernorm(
            pe_reads,
            op.join(reads_dir, sample + "_normalized_pe.fastq.gz"),
            config.assemble.norm_opts,
            config.assemble.cores,
        )
        next_pe = normed_reads
        # run_stats['normalizer'] = apps.kmernorm_version()
        # run_stats['normalization_parameters'] = '"%s"' % config.assemble.norm_opts
        # over 2 due to the reads being interleaved
        # run_stats['normalized_read_count'] = fastx.read_count(normed_reads) / 2
        current_read_count = fastx.read_count(normed_reads) / 2
    else:
        next_pe = pe_reads
    if not current_read_count:
        stop_processing("no reads after normalization", run_stats, run_stats_file)
    # contamination filter against bwa indexes
    if config.assemble.reference:
        # this should likely be a loop
        # for ref_fasta in config.assemble.reference.split(","):
        #     # needs to accommodate temp output file
        #     cfilter_reads = apps.reference_filter(next_pe, ref_fasta,
        #                                           op.join(reads_dir, sample + "_contamfiltered_pe.fastq.gz"),
        #                                           hits_file=None,
        #                                           threshold=config.assemble.reference_threshold,
        #                                           cores=config.assemble.cores)
        cfilter_reads = apps.reference_filter(
            next_pe,
            config.assemble.reference,
            op.join(reads_dir, sample + "_contamfiltered_pe.fastq.gz"),
            hits_file=None,
            threshold=config.assemble.reference_threshold,
            cores=config.assemble.cores,
        )
        # loop ends here
        # run_stats['contamination_reference'] = ";".join(
        #     op.basename(i) for i in config.assemble.reference.split(",")
        # )
        # run_stats['contamination_identity_threshold'] = "%0.02f%%" % (
        #     100 - (100 * config.assemble.reference_threshold)
        # )
        # over 2 due to the reads being interleaved
        # run_stats['passed_contamination_filter'] = fastx.read_count(cfilter_reads) / 2
        current_read_count = fastx.read_count(cfilter_reads) / 2
        next_pe = cfilter_reads
    if not current_read_count:
        stop_processing(
            "no reads after contamination filtering", run_stats, run_stats_file
        )
    # assemble
    if config.assemble.assembly_opts:
        # run_stats['assembler'] = apps.spades_version()
        # run_stats['assembly_parameters'] = '"%s"' % config.assemble.assembly_opts
        # run_stats['passing_contig_length'] = config.assemble.passing_length
        all_contigs = apps.spades(
            sample,
            next_pe,
            op.join(qc_dir, sample + "_all_contigs.fasta"),
            config.assemble.assembly_opts,
            config.assemble.cores,
        )
        if utils.file_exists(all_contigs):
            stats = fastx.fasta_stats(all_contigs)
            # run_stats['all_contigs_count'] = stats[2]
            # run blast against all_contigs
            if config.assemble.blast_db and config.assemble.blast_opts:
                blast_tabular = []
                for blast_db in config.assemble.blast_db.split(","):
                    db_name = os.path.basename(blast_db)
                    blast_archive = apps.blastn(
                        all_contigs,
                        op.join(
                            blast_dir, sample + "_" + db_name + "_all_contigs.ASN1"
                        ),
                        blast_db,
                        config.assemble.blast_opts,
                        config.assemble.cores,
                    )
                    blast_html = apps.blast_formatter(
                        blast_archive,
                        op.join(
                            blast_dir, sample + "_" + db_name + "_all_contigs.html"
                        ),
                        "-html -max_target_seqs 10",
                    )
                    blast_tabular_db = apps.blast_formatter(
                        blast_archive,
                        op.join(blast_dir, sample + "_" + db_name + "_all_contigs.tsv"),
                        "-max_target_seqs 10",
                        (
                            "6 qseqid sseqid pident length mismatch gapopen qstart "
                            "qend sstart send evalue bitscore sallseqid score nident "
                            "positive gaps ppos qframe sframe qseq sseq qlen slen "
                            "salltitles"
                        ),
                    )
                    blast_tabular.append(blast_tabular_db)
            # length filter the contigs
            l_pass, l_fail = fastx.length_filter(
                all_contigs,
                op.join(qc_dir, sample + "_length_passing_contigs.fasta"),
                op.join(qc_dir, sample + "_length_failing_contigs.fasta"),
                length=config.assemble.passing_length,
            )
            if utils.file_exists(l_pass):
                stats = fastx.fasta_stats(l_pass)
                # run_stats['length_passing_contigs_count'] = stats[2]
                # trim then run blast contamination filter
                if (
                    config.assemble.contamination_terms
                    and config.assemble.blast_db
                    and config.assemble.blast_opts
                ):
                    raw_trimmed_contigs = fastx.trim_fastx(
                        l_pass,
                        op.join(qc_dir, sample + "_raw_trimmed_contigs.fasta"),
                        config.assemble.left_trim,
                        config.assemble.right_trim,
                    )
                    final_contigs = apps.blast_results_filter(
                        blast_tabular,
                        raw_trimmed_contigs,
                        op.join(output_dir, sample + "_contigs.fasta"),
                        [
                            i.strip()
                            for i in config.assemble.contamination_terms.split(",")
                        ],
                        contaminated_contigs=op.join(
                            qc_dir, sample + "_contaminated_contigs.fasta"
                        ),
                        length=config.assemble.contamination_length,
                        identity=config.assemble.contamination_identity,
                    )
                    tmp_stats = fastx.fasta_stats(final_contigs)
                    # run_stats['contamination_databases'] = ";".join(
                    #     [
                    #         os.path.basename(i)
                    #         for i in config.assemble.blast_db.split(",")
                    #     ]
                    # )
                    # run_stats['contaminated_contigs'] = run_stats[
                    #     'length_passing_contigs_count'
                    # ] - tmp_stats[
                    #     2
                    # ]
                    # run_stats[
                    #     'contamination_terms'
                    # ] = '"%s"' % config.assemble.contamination_terms
                    # run_stats[
                    #     'contamination_alignment_length'
                    # ] = config.assemble.contamination_length
                    # run_stats[
                    #     'contamination_identity_threshold'
                    # ] = config.assemble.contamination_identity
                    # BLAST has filtered everything
                    if fastx.read_count(final_contigs) == 0:
                        stop_processing(
                            "no contigs survived blast results filtering",
                            run_stats,
                            run_stats_file,
                        )
                # or just trim the contigs by length
                else:
                    final_contigs = fastx.trim_fastx(
                        l_pass,
                        op.join(output_dir, sample + "_contigs.fasta"),
                        config.assemble.left_trim,
                        config.assemble.right_trim,
                    )
                stats = fastx.fasta_stats(final_contigs)
                # FIXME
                if stats[0] < 20000:
                    # run_stats['final_contigs_count'] = 0
                    run_stats["final_assembly_length"] = "NA"
                    run_stats["max_contig_length"] = 0
                    run_stats["gc_content"] = "NA"
                    # FIXME
                    stop_processing(
                        "assembled length less than 20kbp threshold",
                        run_stats,
                        run_stats_file,
                    )
                    os.remove(final_contigs)
                else:
                    # run_stats['final_contigs_count'] = stats[2]
                    run_stats["final_assembly_length"] = stats[0]
                    run_stats["max_contig_length"] = stats[1]
                    run_stats["gc_content"] = stats[3]
                # if run_stats['final_contigs_count'] > 0:
                if stats[2] > 0:
                    # completeness, number_unique_markers, number_multi_copy, taxonomy_contained, taxonomy_sister_lineage
                    checkm_vals = apps.run_checkm(
                        final_contigs, checkm_dir, config.assemble.cores
                    )
                    run_stats["checkM_estimated_completeness"] = checkm_vals[0]
                    # run_stats['unique_marker_genes'] = checkm_vals[1]
                    # run_stats['number_multi_copy'] = checkm_vals[2]
                    # run_stats['checkM_classification'] = ";".join(
                    #     [checkm_vals[3], checkm_vals[4]]
                    # )
                    # functional annotation on final contigs
                    prokka_files = apps.run_prokka(
                        final_contigs,
                        prokka_dir,
                        sample,
                        proteins=config.assemble.prokka_proteins,
                        protein_out=op.join(
                            swissprot_dir, sample + "_prokka-swissprot.tsv"
                        ),
                        cores=config.assemble.cores,
                    )

                    prokka_merged_proteins = op.join(
                        swissprot_dir, sample + "_prokka-swissprot.tsv"
                    )
                    if utils.file_exists(prokka_merged_proteins):
                        cds_stats = apps.parse_prokka_output(prokka_merged_proteins)
                        run_stats["total_CDS_count"] = cds_stats["count"]
                        run_stats["hypothetical_cds_fraction"] = cds_stats[
                            "hypothetical_fraction"
                        ]
                        run_stats["average_cds_length"] = cds_stats["average_length"]
                    run_stats["number_multi_copy"] = checkm_vals[2]
                    run_stats["checkM_classification"] = ";".join(
                        [checkm_vals[3], checkm_vals[4]]
                    )
                    if with_interproscan:
                        apps.run_interproscan(
                            final_contigs,
                            op.join(interproscan_dir, sample + "_interproscan.tsv"),
                            cores=config.assemble.cores,
                        )
                    if (
                        config.assemble.ssu_blast_db
                        and config.assemble.ssu_blast_map
                        and config.assemble.ssu_blast_tre
                    ):
                        ssu_tax = classifier.classify_by_ssu(
                            final_contigs,
                            config.assemble.ssu_blast_db,
                            config.assemble.ssu_blast_map,
                            config.assemble.ssu_blast_tre,
                            op.join(ssu_dir, sample + "_SSU_blast.tsv"),
                            op.join(ssu_dir, sample + "_SSU.fasta"),
                            op.join(ssu_dir, sample + "_SSU.tsv"),
                            min_score=100,
                            top_fraction=0.98,
                            min_length=ssu_min_length,
                            threads=config.assemble.cores,
                        )
                        if isinstance(ssu_tax, six.string_types):
                            run_stats["SSU_classification_1"] = ssu_tax
                            # run_stats['SSU_contig_1'] = "NA"
                        else:
                            for i, (name, length, annotation) in enumerate(
                                ssu_tax, start=1
                            ):
                                run_stats["SSU_classification_%d" % i] = annotation
                                # run_stats['SSU_contig_%d' % i] = "%s (%d bp)" % (
                                # name, length
                                # )
                # run tetramer PCA
                # if config.assemble.pca_script and run_stats['final_contigs_count'] > 1:
                if config.assemble.pca_script and stats[2] > 1:
                    pca_status = apps.tetramer_pca(
                        final_contigs,
                        pca_dir,
                        config.assemble.pca_script,
                        config.assemble.pca_opts,
                        config.assemble.cores,
                    )
                    if not pca_status:
                        run_stats["notes"] = "Tetramer PCA failed"
            else:
                run_stats["notes"] = "no contigs passed length filter"
        else:
            run_stats["notes"] = "assembly failed"
    if "notes" not in run_stats:
        run_stats["notes"] = ""
    print_stats(run_stats_file, run_stats)
    log.info("Assembly of %s is complete." % sample)


def make_archive(tar, file_tuples, overwrite=False):
    """
    tar - name of archive file
    file_tuples - list of (input_path, output_path) tuples
    overwrite - clobber existing tars
    """
    log = logging.getLogger(__name__)
    if op.exists(tar):
        if op.getsize(tar) == 0 or overwrite:
            os.remove(tar)
        else:
            log.info("%s already exists." % tar)
            return tar

    utils.safe_makedir(op.dirname(tar))
    with tarfile.open(tar, "w:gz") as tf:
        for fin, fout in file_tuples:
            if not op.exists(fin) or op.getsize(fin) == 0:
                log.warn("Skipping missing file %s of archive %s" % (fin, tar))
            else:
                tf.add(fin, arcname=fout)
    return tar


def process_sample(sample_meta, input_dir, output_dir, overwrite):
    """
    sample_meta - (customer id, run id, sample name)
    input_dir
    output_dir
    overwrite
    """
    log = logging.getLogger(__name__)
    customer, run_id, sample = sample_meta
    customer = customer.strip("/")
    # assert run_id.count("_") >= 3
    log.info("Processing sample %s of %s" % (sample, run_id))
    files = [
        (
            op.join(input_dir, run_id, sample, sample + "_WGS_QC"),
            op.join(sample, sample + "_WGS_QC"),
        ),
        (
            op.join(input_dir, run_id, sample, sample + "_functional_annotation"),
            op.join(sample, sample + "_functional_annotation"),
        ),
        (
            op.join(input_dir, run_id, sample, sample + "_assembly_info.csv"),
            op.join(sample, sample + "_assembly_info.csv"),
        ),
        (
            op.join(input_dir, run_id, sample, sample + "_contigs.fasta"),
            op.join(sample, sample + "_contigs.fasta"),
        ),
        (
            op.join(
                input_dir,
                run_id,
                sample,
                sample + "_WGS_reads",
                sample + "_pe.fastq.gz",
            ),
            op.join(sample, sample + "_WGS_reads", sample + "_pe.fastq.gz"),
        ),
    ]
    if output_dir.startswith("remote:"):
        with utils.tmp_dir() as tdir:
            dst = "{output_dir}/{customer}/{run_id}/{sample}.tar.gz".format(
                output_dir=output_dir.rstrip("/"),
                customer=customer,
                run_id=run_id,
                sample=sample,
            )
            tar = make_archive(
                op.join(tdir, run_id, customer, sample + ".tar.gz"), files, overwrite
            )
            if utils.cp(tar, dst):
                log.info("archive uploaded")
            else:
                log.error("failed to upload to remote {dst}".format(dst=dst))
    else:
        tar = make_archive(
            op.join(output_dir, run_id, customer, sample + ".tar.gz"), files, overwrite
        )
        log.info("archive created for %s" % sample)
    return tar


@cli.command("deliverables", short_help="create archives from results")
@click.argument("csvfile", type=click.Path(exists=True))
@click.option("--cores", type=int)
@click.option("--overwrite", is_flag=True, help="overwrite existing archives")
@click_config.wrap(module=config, sections=["deliverables", "shared"])
def scgc_deliverables(csvfile, cores, overwrite):
    """
    Takes CSV file with customer, runid, sampleid and creates archives
    in the deliverables folder.

    Setting deliverables to a remote location will trigger an upload, e.g.

    \b
        deliverables: remote:scgc_drive/deliverables/
    """
    log = utils.logger(log_name="Deliverables")
    if not cores:
        cores = config.deliverables.cores
    with open(csvfile, "rU") as fh:
        csvreader = csv.reader(fh)
        samples_meta = [i for i in csvreader]
    log.info("The input file includes %d samples" % len(samples_meta))
    p = multiprocessing.Pool(cores)
    archives = parmap.map(
        process_sample,
        samples_meta,
        config.shared.results,
        config.shared.deliverables,
        overwrite,
        pool=p,
    )
    log.info("Successfully created %d archives" % len(archives))
    out_dirs = [op.dirname(i) for i in archives]
    out_dirs = set(out_dirs)
    log.info("It's recommended that destination directories exist prior to uploading.")
    log.info("To share deliverables, upload the following:")
    for d in out_dirs:
        log.info("rclone sync -v %s remote:scgc_drive/deliverables/etc." % d)


def files_to_df(files):
    df = None
    for f in files:
        if df is None:
            df = pandas.read_csv(f)
            continue

        tdf = pandas.read_csv(f)
        df = pandas.concat([df, tdf])
        del tdf
    return df


@cli.command(
    "compile-stats", short_help="create data table for assembly stats across run IDs"
)
@click.argument("runs", nargs=-1)
@click.option(
    "--out-dir",
    help="override config output directory; prefix path before plate or run ID",
)
@click.option(
    "--out-file", default=None, help="if specified, STDOUT is redirected into this file"
)
@click_config.wrap(module=config, sections=["shared"])
def scgc_compile_stats(runs, out_dir, out_file):
    """
    Create a data table given a run ID or multiple run IDs.
    """
    log = utils.logger(log_name="Stats")
    # stats files from the assembly pipeline
    stats_files = []
    for run in runs:
        if out_dir:
            p = op.join(out_dir, run)
        else:
            p = op.join(config.shared.results, run)
        log.info("Checking %s for stats files." % run)
        if op.exists(p):
            stats_files.extend(glob(op.join(p, "*", "*_*_info.csv")))
        else:
            log.info("Skipping %s as it does not exist." % run)
    if len(stats_files) == 0:
        log.error("No stats files were found.")
        sys.exit(1)
    log.info("Found %d total files." % len(stats_files))
    # take header with most values
    cols = []
    for f in stats_files:
        with open(f) as fh:
            h = fh.readline()
            c = h.strip().split(",")
            if len(c) > len(cols):
                cols = c
    df = files_to_df(stats_files)
    df.sort_values(by=["wgs_run_id", "well"]).to_csv(
        sys.stdout if not out_file else out_file,
        na_rep="NA",
        index=False,
        quoting=csv.QUOTE_NONNUMERIC,
        columns=cols,
    )


@cli.command("coassemble-sample", short_help="coassemble a sample across multiple runs")
@click.argument("output")
@click.argument("samples", nargs=-1)
@click.option(
    "--concatenated-prefix",
    default=None,
    help="optional prefix for concatenated R1 and R2 across runs; should be everything up to _R1.fastq.gz",
)
@click.option(
    "--coassembly-id",
    default="coassembly",
    show_default=True,
    help="affects `scgc assemble` output location",
)
@click.option(
    "--with-interproscan",
    is_flag=True,
    help="run interproscan across genes after assembly",
)
@click_config.wrap(module=config, sections=["assemble", "shared"])
def scgc_coassemble(
    output, samples, concatenated_prefix, coassembly_id, with_interproscan
):
    """
    Samples are defined as run_id:sample_id pairs on the command line and more
    than one is required.
    """
    if len(samples) < 2:
        sys.exit("Specify more than one sample.")
    log = utils.logger(log_name="coassemble")
    r1_paths = []
    r2_paths = []
    for sample in sorted(samples):
        run_id, sample_id = sample.split(":")
        r1 = "{results}/{run}/{sample}/{sample}_WGS_reads/{sample}_trimmed_R1.fastq.gz".format(
            results=config.shared.results,
            run=run_id,
            sample=sample_id,
        )
        if not op.exists(r1):
            log.critical("%s does not exist. Exiting." % r1)
            sys.exit(1)
        r2 = r1.replace("_R1.fastq.gz", "_R2.fastq.gz")
        if not op.exists(r2):
            log.critical("%s does not exist. Exiting." % r2)
            sys.exit(1)
        r1_paths.append(r1)
        r2_paths.append(r2)
    log.info("Using:\n%s" % "\n".join(r1_paths + r2_paths))
    if concatenated_prefix:
        prefix_dir = os.path.dirname(concatenated_prefix)
        utils.safe_makedir(prefix_dir)
        with open(os.path.join(prefix_dir, "info.txt"), "w") as fh:
            print(*samples, sep="\n", file=fh)
        result_r1 = os.path.join("%s_R1.fastq.gz" % concatenated_prefix)
        result_r2 = os.path.join("%s_R2.fastq.gz" % concatenated_prefix)
    else:
        utils.safe_makedir(output)
        with open(os.path.join(output, "info.txt"), "w") as fh:
            print(*samples, sep="\n", file=fh)
        result_r1 = os.path.join(output, "%s_R1.fastq.gz" % os.path.basename(output))
        result_r2 = os.path.join(output, "%s_R2.fastq.gz" % os.path.basename(output))
    log.info("Writing:\n%s\n%s" % (result_r1, result_r2))
    if not os.path.exists(result_r1):
        utils.run("cat %s > %s" % (" ".join(r1_paths), result_r1))
    if not os.path.exists(result_r2):
        utils.run("cat %s > %s" % (" ".join(r2_paths), result_r2))
    utils.run(
        "scgc assemble %s %s %s --output=%s"
        % (
            "--with-interproscan" if with_interproscan else "",
            result_r1,
            coassembly_id,
            output,
        )
    )


@cli.command("join", short_help="cleanup and join PE reads")
@click.argument("r1", type=click.Path(exists=True))
@click.argument("run_id")
@click_config.wrap(module=config, sections=["join", "shared"])
def scgc_join(r1, run_id):
    log = utils.logger(log_name="Join")
    r1 = op.abspath(r1)
    r2 = r1.replace("_R1", "_R2")
    if not op.exists(r2):
        stop_processing("File Not Found: %s" % r2, exit_code=1)
    sample = op.basename(r1).partition("_")[0]
    out_dir = op.join(config.shared.results, run_id, sample)
    utils.safe_makedir(out_dir)
    job_log = op.join(out_dir, "%s_join_info.csv" % sample)
    job_stats = OrderedDict({"sample": sample})
    job_stats["run_id"] = run_id
    job_stats["raw_read_count"] = fastx.read_count(r1)
    current_read_count = job_stats["raw_read_count"]
    if current_read_count < 200:
        stop_processing("too few starting reads", job_stats, job_log)
    qc_dir = utils.safe_makedir(op.join(out_dir, sample + "_QC"))
    reads_dir = utils.safe_makedir(op.join(out_dir, sample + "_reads"))
    blast_dir = op.join(qc_dir, sample + "_joined_reads_blastn")
    if not config.join.skip_qc:
        fastqc = apps.run_fastqc(
            [r1, r2], op.join(qc_dir, sample + "_fastqc"), config.join.cores
        )
    if config.join.trim_opts:
        out_files = apps.trimmomatic(
            [r1, r2],
            [
                op.join(reads_dir, sample + "_trimmed_R1.fastq.gz"),
                op.join(reads_dir, sample + "_orphan_R1.fastq.gz"),
                op.join(reads_dir, sample + "_trimmed_R2.fastq.gz"),
                op.join(reads_dir, sample + "_orphan_R2.fastq.gz"),
            ],
            config.join.trim_opts,
            config.join.cores,
        )
        job_stats["trimmer"] = apps.trimmomatic_version()
        job_stats["trim_parameters"] = '"%s"' % config.join.trim_opts
        job_stats["trimmed_read_count"] = current_read_count = fastx.read_count(
            out_files[0]
        )
        next_r1 = out_files[0]
        next_r2 = out_files[2]
    else:
        next_r1 = r1
        next_r2 = r2
    if not current_read_count:
        stop_processing("no reads survived trimming", job_stats, job_log)
    pe_reads, complexity_read_count = fastx.complexity_filter(
        next_r1,
        next_r2,
        op.join(reads_dir, sample + "_compf_pe.fastq.gz"),
        config.join.complexity_threshold,
        config.join.cores,
    )
    job_stats["complexity_filter_threshold"] = "%0.02f%%" % (
        config.join.complexity_threshold * 100
    )
    job_stats["passed_complexity_filtered"] = current_read_count = complexity_read_count
    if not current_read_count:
        stop_processing("no reads survived complexity filter", job_stats, job_log)
    if config.join.references and len(list(config.join.references.keys())) > 0:
        job_stats["contamination_identity_threshold"] = "%0.02f%%" % (
            100 - (100 * config.assemble.reference_threshold)
        )
        for i, (ref, meta) in enumerate(config.join.references.items()):
            if i > 0 and not current_read_count:
                stop_processing(
                    "all reads were filtered out as contamination", job_stats, job_log
                )
            if meta["save_filtered"]:
                parts = op.basename(pe_reads).split(".")
                # should we incorporate reference name into the pe reads file?
                hits_file = ".".join(
                    ["%s/%s_%s" % (op.dirname(pe_reads), parts[0], ref)] + parts[1:]
                )
                pe_reads, hits_file = apps.reference_filter(
                    pe_reads,
                    meta["prefix"],
                    op.join(reads_dir, sample + "_contf_pe.fastq.gz"),
                    hits_file=hits_file,
                    threshold=config.join.reference_threshold,
                    cores=config.join.cores,
                )
            else:
                pe_reads = apps.reference_filter(
                    pe_reads,
                    meta["prefix"],
                    op.join(reads_dir, sample + "_contf_pe.fastq.gz"),
                    threshold=config.join.reference_threshold,
                    cores=config.join.cores,
                )
            job_stats["passed_%s_filter" % ref] = current_read_count = (
                fastx.read_count(pe_reads) / 2
            )
    if not current_read_count:
        stop_processing("no reads survived contamination filter", job_stats, job_log)
    # could encode params into file name to simplify parameter sweepss
    joined_reads = apps.run_bbmerge(
        pe_reads,
        op.join(reads_dir, sample + "_joined.fastq.gz"),
        k=40,
        extend2=60,
        iterations=5,
        loose="t",
        qtrim2="t",
    )
    job_stats["joined_reads"] = fastx.read_count(joined_reads)
    current_read_count = job_stats["joined_reads"]
    if not current_read_count:
        stop_processing("no reads were joined", job_stats, job_log)
    if "notes" not in job_stats:
        job_stats["notes"] = "Completed Successfully"
    print_stats(job_log, job_stats)


@cli.command("archive", short_help="create a BCL archive")
@click.argument("run_id")
@click.option("--overwrite", is_flag=True, help="overwrite existing archive")
@click_config.wrap(module=config, sections=["shared"])
def archive(run_id, overwrite):
    """Does not perform any integrity checks on the run folder.

    Overwrite to force a new archive to be built.
    """
    log = utils.logger(log_name="Archive")
    run_data = op.join(config.shared.data, run_id)
    archive = op.join(config.shared.archive, run_id + ".tar.gz")
    if op.exists(archive) and not overwrite:
        log.error(
            "An archive already exists for %s; use --overwrite to force rebuilding"
            % run_id
        )
        sys.exit(1)
    if not op.exists(run_data):
        log.error("%s does not exist" % run_data)
        sys.exit(1)
    with utils.file_transaction(archive) as tx:
        cmd = (
            "tar -zcf %s --exclude='*.fastq.gz' --exclude='*.count' --exclude='SAMPLES' --exclude='bcl2fastq.log' %s"
            % (tx, run_data)
        )
        utils.run(cmd, description="Archiving %s" % run_data)
    log.info("Archiving complete for %s" % run_id)


@cli.command("restore", short_help="restore existing BCL archive")
@click.argument("run_id")
@click.option("--overwrite", is_flag=True, help="write into existing directory")
@click_config.wrap(module=config, sections=["shared"])
def restore(run_id, overwrite):
    """If an archive exists for RUN_ID, extract it to shared data directory.

    Overwrite will extract data into an existing folder.
    """
    log = utils.logger(log_name="Restore")
    out_dir = op.join(config.shared.data, run_id)
    archive = op.join(config.shared.archive, run_id + ".tar.gz")
    if not op.exists(archive):
        log.error("No archive exists for %s; checked %s" % (run_id, archive))
        sys.exit(1)
    if op.exists(out_dir) and not overwrite:
        log.error("%s exists and we're not overwriting" % out_dir)
        sys.exit(1)
    cmd = "tar -zxf %s --directory %s" % (archive, config.shared.data)
    utils.run(cmd, description="Extracting %s" % archive)
    log.info("Archive restored for %s" % run_id)


@cli.command("remove-archived", short_help="remove directory if archive exists")
@click.argument("dir_loc")
@click.option("--dev", is_flag=True)
def scgc_remove_archived(dir_loc, dev):
    """removes directory if archive exists, archive location currently hard-coded.
    Args:
        dir_loc (path): path to directory to be deleted
    """

    arch = op.join("/mnt/scgc/simon/archive/", "{}.tar.gz".format(op.basename(dir_loc)))
    assert op.exists(
        arch
    ), "Archive does not exist for {}, make sure all files are archived before attempting to delete.".format(
        dir_loc
    )

    if op.exists(arch):
        if dev:
            print(
                "If dev had been 'False', would be DELETING {} because {} exists".format(
                    dir_loc, arch
                )
            )
        else:
            print("DELETING {} because {} exists".format(dir_loc, arch))
            shutil.rmtree(dir_loc)


@cli.command(
    "classify-seqs", short_help="classify short sequences using SCGC SAGs reference"
)
@click.argument("fastx", type=click.Path(exists=True, resolve_path=True))
@click.argument("db", type=click.Path(exists=True, resolve_path=True))
@click.argument("crestnodes", type=click.Path(exists=True, resolve_path=True))
@click.argument("ncbinodes", type=click.Path(exists=True, resolve_path=True))
@click.argument("outhits", type=click.Path(exists=False, resolve_path=True))
@click.argument("outtsv", type=click.Path(exists=False, resolve_path=True))
@click.option("--min-pid", type=float, default=90)
@click.option("--min-length", type=int, default=100)
@click.option("--top-fraction", type=float, default=0.90)
@click.option("--cores", type=int, default=1)
def run_classify_reads(
    fastx,
    db,
    crestnodes,
    ncbinodes,
    outhits,
    outtsv,
    min_pid,
    min_length,
    top_fraction,
    cores,
):
    """
    Run bbmap using the input FAST(A,Q) against an annotated FASTA database and
    determine the least common taxonomic ancestor among HSPs.
    """
    log = utils.logger(log_name="Classify-SAG")
    log.info("Alignment results are being written to: %s" % outhits)
    log.info("Annotations are being written to: %s" % outtsv)
    classifier.classify_reads(
        fastx,
        db,
        crestnodes,
        ncbinodes,
        outhits,
        outtsv,
        min_pid,
        min_length,
        top_fraction,
        cores,
    )


def samples_from_samplesheet(path):
    samples = []
    start = False
    with open(path, "rU") as fh:
        for line in fh:
            toks = line.strip().split(",")
            if not start:
                # table header processing
                if toks[0] == "Sample_ID":
                    start = True
            elif toks[0]:
                # convert underscores to dashes
                samples.append(toks[0].replace("_", "-").replace(".", "-"))
            else:
                break

    return samples


@cli.command(
    "easy", short_help="run bcl2fastq, assemblies, and stats for a sequence run"
)
@click.argument("run_id")
@click.option(
    "--conversion-threads",
    type=int,
    default=40,
    show_default=True,
    help="bcl2fastq processing threads",
)
@click.option("--subsample", type=int, help="number of reads")
@click.option("--queue", help="qsub queue")
@click.option(
    "--skip-assembly", is_flag=True, help="convert BCL to FASTQ and archive only"
)
@click_config.wrap(module=config, sections=["assemble", "shared"])
def scgc_easy(run_id, conversion_threads, subsample, queue, skip_assembly):
    """Process NextSeq runs through stats compilation. To simultaneously run
    sub-sampled sequences, use --subsample.
    """
    if queue:
        config.assemble.queue = queue
    log = utils.logger(log_name="Easy")
    pfx = op.join(config.shared.data, run_id)
    fq_pfx = op.join(pfx, "Data", "Intensities", "BaseCalls")
    samplesheet = op.join(pfx, "SampleSheet.csv")
    if not op.exists(samplesheet):
        log.error("No SampleSheet was found. Tried %s." % samplesheet)
        sys.exit(1)
    jobs = []
    samples = samples_from_samplesheet(samplesheet)
    if op.exists(op.join(pfx, "bcl2fastq.log")):
        log.info(
            "A log for .BCL conversion exists in the data directory, "
            "so we're skipping that step. To restart entirely, delete "
            "the following:"
        )
        log.info(op.join(pfx, "bcl2fastq.log"))
        # double checking that all fastqs are present
        # for sample in samples:
        #     for idx in ["R1", "R2"]:
        #         assert op.exists(
        #             op.join(
        #                 fq_pfx, "{sample}_{idx}.fastq.gz".format(sample=sample, idx=idx)
        #             )
        #         )
    else:
        cmd = (
            "bcl_to_fastq --determine --overwrite "
            "--processing {threads} "
            "--runfolder-dir {runfolder}"
        ).format(threads=conversion_threads, runfolder=pfx)
        jobs.append(
            utils.submit(
                cmd,
                config.assemble.bcl_queue,
                "bcl-conversion",
                conversion_threads,
                "160:00:00",
                config.assemble.qsub_mem,
                config.assemble.logs,
            )
        )
    conversion = "" if len(jobs) == 0 else jobs[0]
    if not op.exists(op.join(pfx, "{run}_multiqc_report.html".format(run=run_id))):
        report_jobs = []
        cmd = "interop_index-summary {run} --csv=1 > {pfx}/InterOp/index-summary.csv".format(
            run=run_id, pfx=pfx
        )
        report_jobs.append(
            utils.submit(
                cmd,
                queue=config.assemble.queue,
                job_name="bcl-idx-sum",
                cpus=1,
                time="1:00:00",
                log_dir=config.assemble.logs,
                mem="2G",
                depends=[conversion],
            )
        )
        cmd = "interop_summary {run} --csv=1 > {pfx}/InterOp/summary.csv".format(
            run=run_id, pfx=pfx
        )
        report_jobs.append(
            utils.submit(
                cmd,
                queue=config.assemble.queue,
                job_name="bcl-sum",
                cpus=1,
                time="1:00:00",
                mem="2G",
                log_dir=config.assemble.logs,
                depends=[conversion],
            )
        )
        cmd = (
            'multiqc --cl_config "max_table_rows: 2000" --module interop '
            "--no-data-dir --outdir {pfx} --interactive "
            "--title {run} {pfx}"
        ).format(pfx=pfx, run=run_id)
        utils.submit(
            cmd,
            queue=config.assemble.queue,
            job_name="multiqc",
            cpus=1,
            time="1:00:00",
            mem="1G",
            log_dir=config.assemble.logs,
            depends=report_jobs,
        )
    # create a bcl archive
    utils.submit(
        "scgc archive {run}".format(run=run_id),
        config.assemble.queue,
        "bcl-archive",
        config.assemble.qsub_cpus,
        config.assemble.time,
        "2G",
        config.assemble.logs,
        [conversion],
    )

    if skip_assembly:
        log.info(
            "All jobs have been submitted. Per sample assemblies have been skipped."
        )
        return

    for sample in samples:
        cmd = "scgc assemble {r1} {run_id}".format(
            r1=op.join(fq_pfx, "{sample}_R1.fastq.gz".format(sample=sample)),
            run_id=run_id,
        )
        jobs.append(
            utils.submit(
                cmd,
                config.assemble.queue,
                sample,
                config.assemble.qsub_cpus,
                config.assemble.time,
                config.assemble.qsub_mem,
                config.assemble.logs,
                [conversion],
            )
        )
        if subsample:
            cmd = ("scgc assemble --subsample {count} " "{r1} {run_id}").format(
                count=subsample,
                r1=op.join(fq_pfx, "{sample}_R1.fastq.gz".format(sample=sample)),
                run_id=run_id,
            )
            jobs.append(
                utils.submit(
                    cmd,
                    config.assemble.queue,
                    sample,
                    config.assemble.qsub_cpus,
                    config.assemble.time,
                    config.assemble.qsub_mem,
                    config.assemble.logs,
                    [conversion],
                )
            )
    cmd = ("scgc compile-stats --out-file {out_file} " "{run}").format(
        out_file=op.join(
            config.shared.results,
            run_id,
            "{run_id}_assembly_stats.csv".format(run_id=run_id),
        ),
        run=run_id,
    )
    utils.submit(
        cmd,
        config.assemble.queue,
        "compile-stats",
        1,
        "1:00:00",
        "2G",
        config.assemble.logs,
        jobs,
    )
    log.info("All jobs have been submitted. Check status using `qstat`.")


def compile_samples(path):
    samples = set()
    plates = list()
    for fastq_file in glob("%s/*_trimmed_R1.fastq.gz" % path):
        if "_R2" in fastq_file:
            continue
        sample = os.path.basename(fastq_file).partition("_")[0]
        plate = sample.rpartition("-")[0]
        samples.update([sample])
        plates.append(plate)
    return samples, plates


@cli.command(
    "coassemble-plate", short_help="coassemble a plate across multiple sequencing runs"
)
@click.argument("runs", nargs=-1)
@click.option("--with-interproscan", is_flag=True, help="run interproscan across genes")
@click.option("--outdir", help="alternate output location from the config")
@click_config.wrap(module=config, sections=["assemble", "shared", "coassemble"])
def submit_coassembly(runs, with_interproscan, outdir):
    """Specify the runs for LoCoS and the runs of the deep sequencing
    (order independent). Plate IDs are verified before submitting jobs.
    """
    # also verifies all samples of this run are from the same plate
    sample_defs = dict()
    all_plates = list()

    if not outdir:
        outdir = config.coassemble.outdir

    for run in runs:
        # peer into processed data for these reads
        samples, plates = compile_samples(
            os.path.join(config.shared.results, run, "*", "*_WGS_reads")
        )
        all_plates.extend(plates)
        click.echo(
            "Found %d samples in %s across %d plate(s)"
            % (len(samples), run, len(set(plates)))
        )
        sample_defs[run] = samples

    all_samples = Counter()
    for run, samples in six.iteritems(sample_defs):
        for sample in samples:
            all_samples.update([sample])

    intersection_set = []
    either_set = []
    for sample, count in six.iteritems(all_samples):
        if count > 1:
            intersection_set.append(sample)
        else:
            either_set.append(sample)

    click.echo("Preparing to submit %d jobs for coassembly" % len(intersection_set))
    click.echo("and %d jobs for re-assembly" % len(either_set))

    jobs = []

    # coassemblies
    for sample in intersection_set:

        plate = sample.rpartition("-")[0]

        sample_lst = []
        for run, samples in six.iteritems(sample_defs):
            if sample in samples:
                sample_lst.append("%s:%s" % (run, sample))
        sample_str = " ".join(sample_lst)
        prefix = os.path.join(outdir, plate, sample, "%s_WGS_reads" % sample, sample)

        cmd = (
            "scgc coassemble-sample {interproscan} "
            "--coassembly-id {plate} --concatenated-prefix {prefix} "
            "{outdir}/{plate}/{sample} "
            "{samples}"
        ).format(
            interproscan="--with-interproscan" if with_interproscan else "",
            plate=plate,
            prefix=prefix,
            outdir=outdir,
            sample=sample,
            samples=sample_str,
        )

        jobs.append(
            utils.submit(
                cmd,
                queue=config.assemble.queue,
                job_name=sample,
                cpus=config.assemble.qsub_cpus,
                time=config.assemble.time,
                mem=config.assemble.qsub_mem,
                log_dir=config.coassemble.logs,
            )
        )
        # click.echo("run: %s" % cmd)

    # assemblies
    for sample in either_set:

        plate = sample.rpartition("-")[0]

        # figure out which run this sample belongs to
        runid = ""
        for run, samples in six.iteritems(sample_defs):
            if sample in samples:
                runid = run
                # no need to continue checking
                break

        if not runid:
            sys.exit("Unable to determine Run ID for %s" % sample)

        cmd = (
            "scgc assemble {interproscan} " "--output {out_dir} {r1} " "{runid}"
        ).format(
            interproscan="--with-interproscan" if with_interproscan else "",
            out_dir=os.path.join(outdir, plate, sample),
            r1=os.path.join(
                config.shared.results,
                runid,
                "*",
                "*_WGS_reads",
                sample + "_trimmed_R1.fastq.gz",
            ),
            runid=runid,
        )
        jobs.append(
            utils.submit(
                cmd,
                queue=config.assemble.queue,
                job_name=sample,
                cpus=config.assemble.qsub_cpus,
                time=config.assemble.time,
                mem=config.assemble.qsub_mem,
                log_dir=config.assemble.logs,
            )
        )
        # click.echo("run: %s" % cmd)

    for plate in set(all_plates):

        cmd = (
            "scgc compile-stats --out-dir {outdir} "
            "--out-file {outdir}/{plate}_assembly_stats.csv "
            "{plate}"
        ).format(outdir=outdir, plate=plate)
        utils.submit(
            cmd,
            queue=config.assemble.queue,
            job_name="stats-%s" % plate,
            cpus=1,
            time="2:00:00",
            log_dir=config.assemble.logs,
            depends=jobs,
        )
        # click.echo("Stats: %s" % cmd)

    click.echo("All jobs submitted.")


@cli.command("join-all", short_help="run bcl2fastq then bbmerge across all samples")
@click.argument("run_id")
@click.option(
    "--conversion-threads",
    type=int,
    default=40,
    show_default=True,
    help="bcl2fastq processing threads",
)
@click.option("--queue", help="qsub queue")
@click_config.wrap(module=config, sections=["assemble", "shared", "join"])
def scgc_join_all(run_id, conversion_threads, queue):
    if queue:
        config.assemble.queue = queue
    log = utils.logger(log_name="join-all")
    pfx = op.join(config.shared.data, run_id)
    fq_pfx = op.join(pfx, "Data", "Intensities", "BaseCalls")
    samplesheet = op.join(pfx, "SampleSheet.csv")
    if not op.exists(samplesheet):
        log.error("No SampleSheet was found. Tried %s." % samplesheet)
        sys.exit(1)
    jobs = []
    samples = samples_from_samplesheet(samplesheet)
    if op.exists(op.join(pfx, "bcl2fastq.log")):
        log.info(
            "A log for .BCL conversion exists in the data directory, "
            "so we're skipping that step. To restart entirely, delete "
            "the following:"
        )
        log.info(op.join(pfx, "bcl2fastq.log"))
    else:
        cmd = (
            "bcl_to_fastq --determine --overwrite "
            "--processing {threads} "
            "--runfolder-dir {runfolder}"
        ).format(threads=conversion_threads, runfolder=pfx)
        jobs.append(
            utils.submit(
                cmd,
                config.assemble.bcl_queue,
                "bcl-conversion",
                conversion_threads,
                "160:00:00",
                config.assemble.qsub_mem,
                config.assemble.logs,
            )
        )
    conversion = "" if len(jobs) == 0 else jobs[0]
    if not op.exists(op.join(pfx, "{run}_multiqc_report.html".format(run=run_id))):
        report_jobs = []
        cmd = "interop_index-summary {run} --csv=1 > {pfx}/InterOp/index-summary.csv".format(
            run=run_id, pfx=pfx
        )
        report_jobs.append(
            utils.submit(
                cmd,
                queue=config.assemble.queue,
                job_name="bcl-idx-sum",
                cpus=1,
                time="1:00:00",
                log_dir=config.assemble.logs,
                mem="2G",
                depends=[conversion],
            )
        )
        cmd = "interop_summary {run} --csv=1 > {pfx}/InterOp/summary.csv".format(
            run=run_id, pfx=pfx
        )
        report_jobs.append(
            utils.submit(
                cmd,
                queue=config.assemble.queue,
                job_name="bcl-sum",
                cpus=1,
                time="1:00:00",
                mem="2G",
                log_dir=config.assemble.logs,
                depends=[conversion],
            )
        )
        cmd = (
            'multiqc --cl_config "max_table_rows: 2000" --module interop '
            "--no-data-dir --outdir {pfx} --interactive "
            "--title {run} {pfx}"
        ).format(pfx=pfx, run=run_id)
        utils.submit(
            cmd,
            queue=config.assemble.queue,
            job_name="multiqc",
            cpus=1,
            time="1:00:00",
            mem="1G",
            log_dir=config.assemble.logs,
            depends=report_jobs,
        )
    # create a bcl archive
    utils.submit(
        "scgc archive {run}".format(run=run_id),
        config.assemble.queue,
        "bcl-archive",
        config.assemble.qsub_cpus,
        config.assemble.time,
        "2G",
        config.assemble.logs,
        [conversion],
    )

    for sample in samples:
        cmd = "scgc join {r1} {run_id}".format(
            r1=op.join(fq_pfx, "{sample}_R1.fastq.gz".format(sample=sample)),
            run_id=run_id,
        )
        jobs.append(
            utils.submit(
                cmd,
                config.assemble.queue,
                sample,
                1,
                "40:00:00",
                config.assemble.qsub_mem,
                config.assemble.logs,
                [conversion],
            )
        )
    log.info("All jobs have been submitted. Check status using `qstat`.")


@cli.command(
    "complexity-filter", short_help="run nucleotide complexity filter on fastqs"
)
@click.argument("r1")
@click.argument("r2")
@click.argument("pe")
@click.option(
    "--threshold",
    default=0.05,
    type=float,
    help="nucleotide threshold",
)
@click.option(
    "--cores",
    default=1,
    type=int,
    help="cores to utilize",
)
def scgc_complexity_filter(r1, r2, pe, threshold=0.05, cores=1):
    fastx.complexity_filter(r1, r2, pe, threshold, cores)


if __name__ == "__main__":
    cli()



import contextlib
import fnmatch
import itertools
import logging
import parmap
import os
import os.path as op
import shutil
import six
import subprocess
import tempfile
import time
from io import TextIOWrapper
from sarge import capture_stdout, capture_stderr
from six.moves import zip_longest


utils_logger = logging.getLogger(__name__)


def grouper(n, iterable, fillvalue=None):
    args = [iter(iterable)] * n
    return zip_longest(fillvalue=fillvalue, *args)


def multiprocess(f, iterable, *args, **kwargs):
    """
    Map an iterable to a function. Default key function chunks iterable by
    1000s.

    parameters
        f : callable
        iterable : any iterable where each item is sent to f
        args : arguments passed to mapped function
        kwargs : additional arguments for parmap.map

    returns
        mapped function result
    """
    chunksize = kwargs.pop('chunksize', 1000)
    key = kwargs.pop('key', lambda k, l=itertools.count(): next(l)//chunksize)
    for k, g in itertools.groupby(iterable, key=key):
        yield parmap.map(f, g, *args, **kwargs)


def find_files(directory, pattern):
    """
    returns all files (recursively) under directory that match pattern or
    patterns.

    parameters
        directory : where to start searching
        pattern : string pattern(s) to search; wildcards are accepted

    returns
        file paths : list
    """
    found_files = []
    pattern = [pattern] if isinstance(pattern, six.string_types) else pattern
    for root, dirs, files in os.walk(directory):
        for p in pattern:
            for filename in fnmatch.filter(files, p):
                found_files.append(os.path.join(root, filename))
    return found_files


@contextlib.contextmanager
def file_transaction(*rollback_files):
    """
    Wrap file generation in a transaction, moving to output if finishes.
    """
    exts = {".vcf": ".idx", ".bam": ".bai", "vcf.gz": ".tbi", ".fastq.gz": ".count"}
    safe_names, orig_names = _flatten_plus_safe(rollback_files)
    # remove any half-finished transactions
    remove_files(safe_names)
    try:
        if len(safe_names) == 1:
            yield safe_names[0]
        else:
            yield tuple(safe_names)
    # failure -- delete any temporary files
    except:
        remove_files(safe_names)
        remove_tmpdirs(safe_names)
        raise
    # worked -- move the temporary files to permanent location
    else:
        for safe, orig in zip(safe_names, orig_names):
            if os.path.exists(safe):
                shutil.move(safe, orig)
                for check_ext, check_idx in six.iteritems(exts):
                    if safe.endswith(check_ext):
                        safe_idx = safe + check_idx
                        if os.path.exists(safe_idx):
                            shutil.move(safe_idx, orig + check_idx)
        remove_tmpdirs(safe_names)


def remove_tmpdirs(fnames):
    for x in fnames:
        xdir = os.path.dirname(os.path.abspath(x))
        if xdir and os.path.exists(xdir):
            shutil.rmtree(xdir, ignore_errors=True)


def remove_files(fnames):
    for x in fnames:
        if x and os.path.exists(x):
            if os.path.isfile(x):
                os.remove(x)
            elif os.path.isdir(x):
                shutil.rmtree(x, ignore_errors=True)


def _flatten_plus_safe(rollback_files):
    """
    Flatten names of files and create temporary file names.
    """
    tx_files, orig_files = [], []
    for fnames in rollback_files:
        if isinstance(fnames, six.string_types):
            fnames = [fnames]
        for fname in fnames:
            basedir = safe_makedir(os.path.dirname(fname))
            tmpdir = safe_makedir(tempfile.mkdtemp(dir=basedir))
            tx_file = os.path.join(tmpdir, os.path.basename(fname))
            tx_files.append(tx_file)
            orig_files.append(fname)
    return tx_files, orig_files


def safe_makedir(dname):
    """
    Make a directory if it doesn't exist, handling concurrent race conditions.
    """
    if not dname:
        return dname
    num_tries = 0
    max_tries = 5
    while not os.path.exists(dname):
        try:
            os.makedirs(dname)
        except OSError:
            if num_tries > max_tries:
                raise
            num_tries += 1
            time.sleep(2)
    return dname


def file_exists(fnames):
    """
    Check if a file or files exist and are non-empty.

    parameters
        fnames : file path as string or paths as list; if list, all must exist

    returns
        boolean
    """
    if isinstance(fnames, six.string_types):
        fnames = [fnames]
    for f in fnames:
        if not os.path.exists(f) or os.path.getsize(f) == 0:
            return False
    return True


def stdout_iter(cmd):
    p = capture_stdout(cmd)
    for line in TextIOWrapper(p.stdout):
        yield line


def run(cmd, description=None, iterable=None, retcodes=[0]):
    """Runs a command using check_call.

    Args:
        cmd (str): shell command as a string
        description (Optional[str]):
        iterable (Optional[str]):
    """
    if description:
        utils_logger.info(description)
    utils_logger.info("$> %s" % cmd)
    if iterable:
        return stdout_iter(cmd)
    else:
        p = capture_stderr(cmd)
        if p.returncode not in retcodes:
            for line in TextIOWrapper(p.stderr):
                utils_logger.error(line.strip())
            raise subprocess.CalledProcessError(p.returncode, cmd=cmd)
        else:
            return p


def submit(cmd, queue="normal", job_name="scgcpy", cpus=8, time="24:00:00", mem='5G',
           log_dir="logs", depends=[], umask="0002"):
    log_dir = os.path.realpath(os.path.expanduser(log_dir))
    if not os.path.exists(log_dir):
        os.makedirs(log_dir)

    depends = [i for i in depends if i]
    qsub_cmd = ("qsub -N {job_name} -q {queue} "
                "-l walltime={time},ncpus={cpus},mem={mem} "
                "-W umask={umask} -j oe -o {log_dir} "
                "{dependencies}").format(job_name=job_name,
                                         queue=queue,
                                         time=time,
                                         cpus=cpus,
                                         umask=umask,
                                         mem=mem,
                                         log_dir=log_dir,
                                         dependencies="" if not depends else "-W depend=afterany:%s" % ":".join(depends[-50:]))
    cmd = '''echo \"{cmd}\" | {qsub_cmd}'''.format(cmd=cmd, qsub_cmd=qsub_cmd)
    utils_logger.info(cmd)
    p = capture_stdout(cmd, shell=True)
    job_id = ""
    for line in TextIOWrapper(p.stdout):
        job_id = line.strip()
        break
    return job_id


def cp(src, dst):
    """
    will transfer to rclone's 'remote' if either src or dst starts with 'remote:'.
    """
    if src.startswith("remote:") or dst.startswith("remote:"):
        try:
            run("rclone copy {src} {dst}".format(src=src, dst=dst))
        except subprocess.CalledProcessError:
            utils_logger.warn("failed transferring {src} to {dst}".format(src=src, dst=dst))
            return False
    else:
        shutil.copyfile(src, dst)
    return True


def mv(src, dst):
    shutil.move(src, dst)


def logger(log_file=None, log_name=None, log_level=logging.INFO,
           log_format="[%(asctime)s %(name)s %(levelname)s] %(message)s"):
    # dict(info=logging.INFO, debug=logging.DEBUG) and then map to proper level
    logging.basicConfig(level=log_level, format=log_format)
    log = logging.getLogger(log_name)
    if log_file:
        logfile = logging.FileHandler(log_file)
        logfile.setLevel(log_level)
        logfile.setFormatter(logging.Formatter(log_format))
        log.addHandler(logfile)
    return log


@contextlib.contextmanager
def tmp_dir():
    d = None
    try:
        d = tempfile.mkdtemp()
        yield d
    finally:
        if d:
            shutil.rmtree(d)


def is_plain(input_file):
    """Use `file` to verify input is plain text.

    Args:
        input_file (str): path to file

    Returns:
        boolean

    Raises:
        IOError
    """
    if not op.exists(input_file):
        raise IOError(2, "No such file:", input_file)
    for output in run("file -b --mime-type " + input_file, iterable=True):
        if "plain" in output:
            return True
    return False


def pigz_file(fname, cores=1):
    if not op.exists(fname):
        raise IOError(2, "No such file:", fname)
    file_path = ""
    out_file = ""
    if fname.endswith('.gz'):
        out_file = fname
        if is_plain(fname):
            # need to rename the file
            file_path = op.splitext(fname)[0]
            shutil.move(fname, file_path)
    else:
        out_file = fname + '.gz'
        file_path = fname
    if file_path:
        try:
            run("pigz -f --best -p %d %s" % (cores, file_path))
        except:
            run("gzip -f --best %s" % file_path)
    return out_file


def gunzip_file(fname):
    """Decompress fname using `gzip -d`.

    Args:
        fname (str): file path to decompress

    Returns:
        str
    """
    out_file, ext = op.splitext(fname)
    run("gzip -d -f %s" % fname)
    return out_file


def touch(fname, times=None):
    """https://stackoverflow.com/questions/1158076/implement-touch-using-python"""
    with open(fname, 'a'):
        os.utime(fname, times)



import logging
import os
import sys
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns

from .utils import file_exists, run


logger = logging.getLogger(__name__)
sns.set_context('paper')
sns.set_style('whitegrid')


def out_file_format(out_file):
    _, ext = os.path.splitext(out_file)
    if not ext == '.pdf' or ext == '.png':
        raise IOError('Format must be either pdf of png, not "%s"' % ext)
    return ext[1:]


# TODO: move this to stats.py
def Lc(v):
    """
    Replicate the functionality of R's Lc().

    In R:
    > library(ineq)
    > t <- c(7,13,16)
    > lc <- Lc(t)
    > lc$p
    [1] 0.0000000 0.3333333 0.6666667 1.0000000
    > lc$L
    [1] 0.0000000 0.1944444 0.5555556 1.0000000
    > ineq(t, type="Gini")
    [1] 0.1666667

    :type v: list
    :param v: list of numbers

    :rtype: tuple
    :return: x values, y values, Gini coefficient

    >>> t = [7,13,16]
    >>> Lc(t)
    ((0.0, 0.33..., 0.66..., 1.0), (0.0, 0.19..., 0.55..., 1.0), 0.16...)
    """
    v_len = len(v)
    assert v_len > 0

    a = np.array([0.] + v)
    a.sort()
    x = np.arange(0., v_len + 1) / v_len
    y = a.cumsum() / a.sum()
    G = 1 + (1 - 2 * y.sum()) / v_len
    return x, y, G


# TODO: this should be generic; wrap function to get vector under something else
def lorenz_curve(bam, out_file):
    """
    Calculate Gini coefficient and plot Lorenze curve.

    :type bam: string
    :param bam: aligned reads in bam format

    :type out_file: string
    :param out_file: output file path with extension

    :rtype: float
    :return: Gini coefficient
    """
    file_format = out_file_format(out_file)
    coverages = []
    cmd = '| bedtools genomecov -5 -d -ibam %s' % bam
    header = ['name', 'start', 'coverage']
    for toks in reader(cmd, header=header):
        coverages.append(int(toks['coverage']))
    x, y, G = Lc(coverages)
    logger.info('Gini coefficient:', G, file=sys.stderr)

    plt.plot(x, y)
    plt.plot([0, 1], [0, 1])
    plt.xlabel('Fraction of Genome')
    plt.ylabel('Fraction of Reads')
    plt.savefig(out_file, bbox_inches='tight', format=file_format)
    return G


def get_coverage(bam):
    bedgraph = ""
    filename, _ = path.splitext(bam)
    bedgraph = filename + ".bedgraph.gz"
    if not file_exists(bedgraph):
        cmd = ("bedtools genomecov -bga -ibam {bam} "
            "| bedtools sort -i - "
            "| gzip > {bedgraph}").format(bam=bam, bedgraph=bedgraph)
        run(cmd)
    return bedgraph


def genome_coverage(bam, out_file):
    file_format = out_file_format(out_file)
    fig = plt.figure(figsize=(12, 3))
    bg = get_coverage(bam)
    h = ['chrom', 'start', 'stop', 'count']
    x, y = list(zip(*[(int(t['start']), int(t['count'])) for t in reader(bg, header=h)]))
    plt.plot(x, y)
    plt.xlabel('Genomic Position')
    plt.ylabel('Read Count')
    plt.savefig(out_file, bbox_inches='tight', format=file_format)
